/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.sql.*;
import java.util.*;

import j4d.open.*;

/**
*	Represents a single boolean expression from WHERE clause.
*	This is not a very generic class. Only a simple expression is supported:
*	a column name, followed by a comparison operator, followed by a value.
*/
class BooleanExpression
{
	// Column to compare
	private SQLColumnMetaData			m_sqlcMD = null;
	
	// Comparison operators
	static final String					EQUAL = "=";
	static final String					NOT_EQUAL = "!=";
	static final String					LESS_THAN = "<";
	static final String					LESS_THAN_OR_EQUAL = "<=";
	static final String					GREATER_THAN = ">";
	static final String					GREATER_THAN_OR_EQUAL = ">=";
	private String						m_strComparisonOperator = null; // undefined
	
	// Value to compare with
	private Object						m_objValue = null;
	
	// Logical operator that follows this boolean expression
	private String						m_strLogicalOperator = null; // undefined
	
	BooleanExpression ( SQLColumnMetaData sqlcMD, String strComparisonOperator, Object objValue ) throws Exception
	{
		if ( sqlcMD == null )
			throw new Exception (
							"com.fourd.jdbc.BooleanExpression::BooleanExpression: " +
							"Column meta data can not be null" );

		if ( strComparisonOperator == null )
			throw new Exception (
							"com.fourd.jdbc.BooleanExpression::BooleanExpression: " +
							"Comparison operator can not be null" );

		if (	!strComparisonOperator. equals ( EQUAL ) &&
			!strComparisonOperator. equals ( NOT_EQUAL ) &&
			!strComparisonOperator. equals ( LESS_THAN ) &&
			!strComparisonOperator. equals ( LESS_THAN_OR_EQUAL ) &&
			!strComparisonOperator. equals ( GREATER_THAN ) &&
			!strComparisonOperator. equals ( GREATER_THAN_OR_EQUAL ) )
			throw new Exception (
							"com.fourd.jdbc.BooleanExpression::BooleanExpression: " +
							"Unknown comparison operator - " + strComparisonOperator );

		if ( objValue == null )
			throw new Exception (
							"com.fourd.jdbc.BooleanExpression::BooleanExpression: " +
							"Value can not be null" );
							
		m_sqlcMD = sqlcMD;
		m_strComparisonOperator = strComparisonOperator. toUpperCase ( );
		m_objValue = objValue;
	}
	
	SQLColumnMetaData getColumnMetaData ( )
	{
		return m_sqlcMD;
	}
	
	String getLogicalOperator ( )
	{
		return m_strLogicalOperator;
	}
	
	void setLogicalOperator ( String strLogicalOperator ) throws SQLException
	{
		if ( strLogicalOperator == null )
		{
			m_strLogicalOperator = null;
			/*throw new SQLException (
								"com.fourd.jdbc.BooleanExpression::setLogicalOperator: " +
								"Logical operator can not be null" );*/
								
			return;
		}
			
		if (	!strLogicalOperator. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_AND ) &&
			!strLogicalOperator. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_OR ) )
			throw new SQLException (
							"com.fourd.jdbc.BooleanExpression::setLogicalOperator: " +
							"Unknown logical operator - " + strLogicalOperator );

		m_strLogicalOperator = strLogicalOperator. toUpperCase ( );
	}
	
	byte getopLogicalOperator ( )
	{
		if ( m_strLogicalOperator == null )
			return opConstants. NONE;
			
		if ( m_strLogicalOperator. equals ( SQLCommand. SQL_KEYWORD_AND ) )
			return opConstants. AND;
			
		return opConstants. OR;
	}
	
	/*Object getValue ( )
	{
		return m_objValue;
	}
	*/
	
	/**
	*	Returns this expression's value as opData object.
	*	nType contains opConstant type for resulting opData object.
	*/
	opData getopValue ( int nType, opData [ ] szopdParameters ) throws SQLException
	{
		if ( m_objValue instanceof Parameter )
		{
			Parameter	param = ( Parameter ) m_objValue;
			if ( param. getIndex ( ) >= szopdParameters. length )
				throw new SQLException (
								"com.fourd.jdbc.BooleanExpression::getopValue: " +
								"Index out of bounds. Parameter count - " +
								szopdParameters. length +
								", index - " + param. getIndex ( ) );
			
			opData		opD = szopdParameters [ param. getIndex ( ) ];
			if ( opD == null )
				throw new SQLException (
								"com.fourd.jdbc.BooleanExpression::getopValue: " +
								"Parameter value is null" );
								
			return opD;
		}
		
		// We know that m_objValue in fact now contains String
		if ( !( m_objValue instanceof String ) )
			throw new SQLException (
								"com.fourd.jdbc.BooleanExpression::Unexpected value class. Expected java.lang.String " +
								"but found" + m_objValue. getClass ( ). getName ( ) );
		
		String	strValue = ( String ) m_objValue;
		if ( nType == opConstants. ALPHANUMERIC )
			return new opData ( nType, strValue );
		else if ( nType == opConstants. BOOLEAN )
			if ( strValue. equals ( SQLCommand. SQL_KEYWORD_TRUE ) )
				return new opData ( nType, true );
			else if ( strValue. equals ( SQLCommand. SQL_KEYWORD_FALSE ) )
				return new opData ( nType, false );
			else
				throw new SQLException (
									"com.fourd.jdbc.BooleanExpression::getopValue: " +
									"Invalid boolean constant - " + strValue );
		else if ( nType == opConstants. INTEGER )
			try
			{
				short	nResult = new Short ( strValue ). shortValue ( );
				return new opData ( nType, nResult );
			}
			catch ( NumberFormatException nfE )
			{
				throw new SQLException (
									"com.fourd.jdbc.BooleanExpression::getopValue: " +
									"Invalid short value - " + strValue + 
									 "\n" + nfE. getMessage ( ) );
			}
		else if ( nType == opConstants. LONGINTEGER )
			try
			{
				int		nResult = new Integer ( strValue ). intValue ( );
				return new opData ( nType, nResult );
			}
			catch ( NumberFormatException nfE )
			{
				throw new SQLException (
									"com.fourd.jdbc.BooleanExpression::getopValue: " +
									"Invalid integer value - " + strValue + 
									 "\n" + nfE. getMessage ( ) );
			}
		else if ( nType == opConstants. REAL )
			try
			{
				double		nResult = new Double ( strValue ). doubleValue ( );
				return new opData ( nType, nResult );
			}
			catch ( NumberFormatException nfE )
			{
				throw new SQLException (
									"com.fourd.jdbc.BooleanExpression::getopValue: " +
									"Invalid real value - " + strValue + 
									 "\n" + nfE. getMessage ( ) );
			}
		else if ( nType == opConstants. TEXT )
			return new opData ( nType, strValue );
		else if ( nType == opConstants. DATE || nType == opConstants. TIME )
		{
			java.util.Date		dtResult = SQLTools. parseDate ( strValue );
			if ( dtResult == null )
				throw new SQLException (
									"com.fourd.jdbc.BooleanExpression::getopValue: " +
									"Invalid date - " + strValue );
			
			GregorianCalendar		gcResult = new GregorianCalendar ( );
			gcResult. setTime ( dtResult );
			
			return new opData ( nType, gcResult );
		}
		else if ( nType == opConstants. BLOB )
			throw new SQLException (
								"com.fourd.jdbc.BooleanExpression::getopValue: " +
								"Use of BLOBs is not allowed in boolean expressions" );
		else if ( nType == opConstants. PICTURE )
			throw new SQLException (
								"com.fourd.jdbc.BooleanExpression::getopValue: " +
								"Unimplemented data type - PICTURE" );
		else if ( nType == opConstants. SUBTABLE )
			throw new SQLException (
								"com.fourd.jdbc.BooleanExpression::getopValue: " +
								"Unimplemented data type - SUBTABLE" );
		else if ( nType == opConstants. ALPHANUMERIC2 )
			throw new SQLException (
								"com.fourd.jdbc.BooleanExpression::getopValue: " +
								"Unimplemented data type - ALPHANUMERIC2" );
			
		throw new SQLException (
							"com.fourd.jdbc.BooleanExpression::getopValue: " +
							"Unknown data type - " + nType );
	}
	
	String getComparisonOperator ( )
	{
		return m_strComparisonOperator;
	}
	
	byte getopComparisonOperator ( ) // throws Exception
	{
		if ( m_strComparisonOperator == null )
			return opConstants. NONE;
			/*throw new SQLException (
								"com.fourd.jdbc.BooleanExpression::getopComparisonOperator: " +
								"Comparison operator is not set" );
			*/
			
		if ( m_strComparisonOperator. equals ( EQUAL ) )
			return opConstants. EQUAL;
		if ( m_strComparisonOperator. equals ( NOT_EQUAL ) )
			return opConstants. NOT_EQUAL;
		if ( m_strComparisonOperator. equals ( LESS_THAN ) )
			return opConstants. LESS_THAN;
		if ( m_strComparisonOperator. equals ( LESS_THAN_OR_EQUAL ) )
			return opConstants. LESS_THAN_OR_EQUAL;
		if ( m_strComparisonOperator. equals ( GREATER_THAN ) )
			return opConstants. GREATER_THAN;
			
		// if ( m_strComparisonOperator. equals ( GREATER_THAN_OR_EQUAL ) )
			return opConstants. GREATER_THAN_OR_EQUAL;
	}
	
	/**
	*	Throws an exception if strComparisonOperator represents unsupported/unknown comparison operator.
	*/
	static void validateComparisonOperator ( String strComparisonOperator ) throws Exception
	{
		if ( strComparisonOperator == null )
			throw new Exception (
							"com.fourd.jdbc.BooleanExpression::validateComparisonOperator: " +
							"Comparison operator is null" );

		if (	!strComparisonOperator. equals ( EQUAL ) &&
			!strComparisonOperator. equals ( NOT_EQUAL ) &&
			!strComparisonOperator. equals ( LESS_THAN ) &&
			!strComparisonOperator. equals ( LESS_THAN_OR_EQUAL ) &&
			!strComparisonOperator. equals ( GREATER_THAN ) &&
			!strComparisonOperator. equals ( GREATER_THAN_OR_EQUAL ) )
			throw new Exception (
							"com.fourd.jdbc.BooleanExpression::validateComparisonOperator: " +
							"Unknown comparison operator - " + strComparisonOperator );
	}
}