/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.sql.*;
import java.util.*;

import j4d.open.*;

class DatabaseMetaDataImpl implements DatabaseMetaData
{
	private static final String			DATABASE_PRODUCT_NAME = "4D Database";
	
	private ConnectionImpl				m_connectionImpl = null;
	// An object that will handle all communication with the DB.
	private opProcess					m_opProcess = null;
	
	// Tables meta data
	private SQLTableMetaData [ ]		m_szsqltMD = null;
	
	private Object						m_objDataAccessLock = null;

	DatabaseMetaDataImpl (
					opProcess opP,
					Object objDataAccessLock,
					ConnectionImpl connectionImpl ) throws SQLException
	{
		m_opProcess = opP;
		m_objDataAccessLock = objDataAccessLock;
		m_connectionImpl = connectionImpl;
		
		lazyLoadMetaData ( );
		// loadMetaData ( );
	}
	
	public boolean allProceduresAreCallable ( ) throws SQLException
	{
		/*
			Not all 4D methods can be called as JDBC stored procedures.
			For security reasons or incompatible input parameter types, etc...
		*/

		return false;
	}

	public boolean allTablesAreSelectable ( ) throws SQLException
	{
		/*
			Can use have no read, no write access?
		*/
		
		return false;
	}

	public String getURL ( ) throws SQLException
	{
		return m_connectionImpl. getHost ( );
	}

	public String getUserName ( ) throws SQLException
	{
		return m_connectionImpl. getUserName ( );
	}

	public boolean isReadOnly ( ) throws SQLException
	{
		return false;
	}

	public boolean nullsAreSortedHigh ( ) throws SQLException
	{
		throwUnsupportedException ( );
		
		return false;
	}

	public boolean nullsAreSortedLow ( ) throws SQLException
	{
		throwUnsupportedException ( );
		
		return false;
	}

	public boolean nullsAreSortedAtStart ( ) throws SQLException
	{
		throwUnsupportedException ( );
		
		return false;
	}

	public boolean nullsAreSortedAtEnd ( ) throws SQLException
	{
		throwUnsupportedException ( );
		
		return false;
	}

	public String getDatabaseProductName ( ) throws SQLException
	{
		return DATABASE_PRODUCT_NAME;
	}

	public String getDatabaseProductVersion ( ) throws SQLException
	{
		throwUnimplementedException ( "getDatabaseProductVersion" );
		
		return null;
	}

	public String getDriverName ( ) throws SQLException
	{
		return DriverImpl. DRIVER_NAME;
	}

	public String getDriverVersion ( ) throws SQLException
	{
		return DriverImpl. VERSION;
	}

	public int getDriverMajorVersion ( )
	{
		return DriverImpl. MAJOR_VERSION;
	}

	public int getDriverMinorVersion ( )
	{
		return DriverImpl. MINOR_VERSION;
	}

	public boolean usesLocalFiles ( ) throws SQLException
	{
		return false;
	}

	public boolean usesLocalFilePerTable ( ) throws SQLException
	{
		return false;
	}

	public boolean supportsMixedCaseIdentifiers ( ) throws SQLException
	{
		return false;
	}

	public boolean storesUpperCaseIdentifiers ( ) throws SQLException
	{
		return false;
	}

	public boolean storesLowerCaseIdentifiers ( ) throws SQLException
	{
		return false;
	}

	public boolean storesMixedCaseIdentifiers ( ) throws SQLException
	{
		return true;
	}

	public boolean supportsMixedCaseQuotedIdentifiers ( ) throws SQLException
	{
		return true;
	}

	public boolean storesUpperCaseQuotedIdentifiers ( ) throws SQLException
	{
		return false;
	}

	public boolean storesLowerCaseQuotedIdentifiers ( ) throws SQLException
	{
		return false;
	}

	public boolean storesMixedCaseQuotedIdentifiers ( ) throws SQLException
	{
		return true;
	}

	public String getIdentifierQuoteString ( ) throws SQLException
	{
		// throwUnimplementedException ( "getIdentifierQuoteString" );
		
		return " ";
	}

	public String getSQLKeywords ( ) throws SQLException
	{
		throwUnimplementedException ( "getSQLKeywords" );
		
		return null;
	}

	public String getNumericFunctions ( ) throws SQLException
	{
		throwUnimplementedException ( "getNumericFunctions" );
		
		return null;
	}

	public String getStringFunctions ( ) throws SQLException
	{
		// throwUnimplementedException ( "getStringFunctions" );
		
		return "";
	}

	public String getSystemFunctions ( ) throws SQLException
	{
		throwUnimplementedException ( "getSystemFunctions" );
		
		return null;
	}

	public String getTimeDateFunctions ( ) throws SQLException
	{
		throwUnimplementedException ( "getTimeDateFunctions" );
		
		return null;
	}

	public String getSearchStringEscape ( ) throws SQLException
	{
		throwUnimplementedException ( "getSearchStringEscape" );
		
		return null;
	}

	public String getExtraNameCharacters ( ) throws SQLException
	{
		throwUnimplementedException ( "getExtraNameCharacters" );
		
		return null;
	}

	public boolean supportsAlterTableWithAddColumn ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsAlterTableWithAddColumn" );
		
		return false;
	}

	public boolean supportsAlterTableWithDropColumn ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsAlterTableWithDropColumn" );
		
		return false;
	}

	public boolean supportsColumnAliasing ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsColumnAliasing" );
		
		return false;
	}

	public boolean nullPlusNonNullIsNull ( ) throws SQLException
	{
		throwUnimplementedException ( "nullPlusNonNullIsNull" );
		
		return false;
	}

	public boolean supportsConvert ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsConvert" );
		
		return false;
	}

	public boolean supportsConvert(int arg0, int arg1) throws SQLException
	{
		throwUnimplementedException ( "supportsConvert" );
		
		return false;
	}

	public boolean supportsTableCorrelationNames ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsTableCorrelationNames" );
		
		return false;
	}

	public boolean supportsDifferentTableCorrelationNames ( )
		throws SQLException
	{
		throwUnimplementedException ( "supportsDifferentTableCorrelationNames" );
		
		return false;
	}

	public boolean supportsExpressionsInOrderBy ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsExpressionsInOrderBy" );
		
		return false;
	}

	public boolean supportsOrderByUnrelated ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsOrderByUnrelated" );
		
		return false;
	}

	public boolean supportsGroupBy ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsGroupBy" );
		
		return false;
	}

	public boolean supportsGroupByUnrelated ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsGroupByUnrelated" );
		
		return false;
	}

	public boolean supportsGroupByBeyondSelect ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsGroupByBeyondSelect" );
		
		return false;
	}

	public boolean supportsLikeEscapeClause ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsLikeEscapeClause" );
		
		return false;
	}

	public boolean supportsMultipleResultSets ( ) throws SQLException
	{
		return false;
	}

	public boolean supportsMultipleTransactions ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsMultipleTransactions" );
		
		return false;
	}

	public boolean supportsNonNullableColumns ( ) throws SQLException
	{
		// throwUnimplementedException ( "supportsNonNullableColumns" );
		
		return false;
	}

	public boolean supportsMinimumSQLGrammar ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsMinimumSQLGrammar" );
		
		return false;
	}

	public boolean supportsCoreSQLGrammar ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsCoreSQLGrammar" );
		
		return false;
	}

	public boolean supportsExtendedSQLGrammar ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsExtendedSQLGrammar" );
		
		return false;
	}

	public boolean supportsANSI92EntryLevelSQL ( ) throws SQLException
	{
		// throwUnimplementedException ( "supportsANSI92EntryLevelSQL" );
		
		return true;
	}

	public boolean supportsANSI92IntermediateSQL ( ) throws SQLException
	{
		// throwUnimplementedException ( "supportsANSI92IntermediateSQL" );
		
		return false;
	}

	public boolean supportsANSI92FullSQL ( ) throws SQLException
	{
		// throwUnimplementedException ( "supportsANSI92FullSQL" );
		
		return false;
	}

	public boolean supportsIntegrityEnhancementFacility ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsIntegrityEnhancementFacility" );
		
		return false;
	}

	public boolean supportsOuterJoins ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsOuterJoins" );
		
		return false;
	}

	public boolean supportsFullOuterJoins ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsFullOuterJoins" );
		
		return false;
	}

	public boolean supportsLimitedOuterJoins ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsLimitedOuterJoins" );
		
		return false;
	}

	public String getSchemaTerm ( ) throws SQLException
	{
		throwUnimplementedException ( "getSchemaTerm" );
		
		return null;
	}

	public String getProcedureTerm ( ) throws SQLException
	{
		throwUnimplementedException ( "getProcedureTerm" );
		
		return null;
	}

	public String getCatalogTerm ( ) throws SQLException
	{
		throwUnimplementedException ( "getCatalogTerm" );
		
		return null;
	}

	public boolean isCatalogAtStart ( ) throws SQLException
	{
		throwUnimplementedException ( "isCatalogAtStart" );
		
		return false;
	}

	public String getCatalogSeparator ( ) throws SQLException
	{
		return "";
	}

	public boolean supportsSchemasInDataManipulation ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsSchemasInDataManipulation" );
		
		return false;
	}

	public boolean supportsSchemasInProcedureCalls ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsSchemasInProcedureCalls" );
		
		return false;
	}

	public boolean supportsSchemasInTableDefinitions ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsSchemasInTableDefinitions" );
		
		return false;
	}

	public boolean supportsSchemasInIndexDefinitions ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsSchemasInIndexDefinitions" );
		
		return false;
	}

	public boolean supportsSchemasInPrivilegeDefinitions ( )
		throws SQLException
	{
		throwUnimplementedException ( "supportsSchemasInPrivilegeDefinitions" );
		
		return false;
	}

	public boolean supportsCatalogsInDataManipulation ( ) throws SQLException
	{
		return false;
	}

	public boolean supportsCatalogsInProcedureCalls ( ) throws SQLException
	{
		return false;
	}

	public boolean supportsCatalogsInTableDefinitions ( ) throws SQLException
	{
		return false;
	}

	public boolean supportsCatalogsInIndexDefinitions ( ) throws SQLException
	{
		return false;
	}

	public boolean supportsCatalogsInPrivilegeDefinitions ( )
		throws SQLException
	{
		return false;
	}

	public boolean supportsPositionedDelete ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsPositionedDelete" );
		
		return false;
	}

	public boolean supportsPositionedUpdate ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsPositionedUpdate" );
		
		return false;
	}

	public boolean supportsSelectForUpdate ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsSelectForUpdate" );
		
		return false;
	}

	public boolean supportsStoredProcedures ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsStoredProcedures" );
		
		return false;
	}

	public boolean supportsSubqueriesInComparisons ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsSubqueriesInComparisons" );
		
		return false;
	}

	public boolean supportsSubqueriesInExists ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsSubqueriesInExists" );
		
		return false;
	}

	public boolean supportsSubqueriesInIns ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsSubqueriesInIns" );
		
		return false;
	}

	public boolean supportsSubqueriesInQuantifieds ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsSubqueriesInQuantifieds" );
		
		return false;
	}

	public boolean supportsCorrelatedSubqueries ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsCorrelatedSubqueries" );
		
		return false;
	}

	public boolean supportsUnion ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsUnion" );
		
		return false;
	}

	public boolean supportsUnionAll ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsUnionAll" );
		
		return false;
	}

	public boolean supportsOpenCursorsAcrossCommit ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsOpenCursorsAcrossCommit" );
		
		return false;
	}

	public boolean supportsOpenCursorsAcrossRollback ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsOpenCursorsAcrossRollback" );
		
		return false;
	}

	public boolean supportsOpenStatementsAcrossCommit ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsOpenStatementsAcrossCommit" );
		
		return false;
	}

	public boolean supportsOpenStatementsAcrossRollback ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsOpenStatementsAcrossRollback" );
		
		return false;
	}

	public int getMaxBinaryLiteralLength ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxBinaryLiteralLength" );
		
		return 0;
	}

	public int getMaxCharLiteralLength ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxCharLiteralLength" );
		
		return 0;
	}

	public int getMaxColumnNameLength ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxColumnNameLength" );
		
		return 0;
	}

	public int getMaxColumnsInGroupBy ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxColumnsInGroupBy" );
		
		return 0;
	}

	public int getMaxColumnsInIndex ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxColumnsInIndex" );
		
		return 0;
	}

	public int getMaxColumnsInOrderBy ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxColumnsInOrderBy" );
		
		return 0;
	}

	public int getMaxColumnsInSelect ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxColumnsInSelect" );
		
		return 0;
	}

	public int getMaxColumnsInTable ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxColumnsInTable" );
		
		return 0;
	}

	public int getMaxConnections ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxConnections" );
		
		return 0;
	}

	public int getMaxCursorNameLength ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxCursorNameLength" );
		
		return 0;
	}

	public int getMaxIndexLength ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxIndexLength" );
		
		return 0;
	}

	public int getMaxSchemaNameLength ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxSchemaNameLength" );
		
		return 0;
	}

	public int getMaxProcedureNameLength ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxProcedureNameLength" );
		
		return 0;
	}

	public int getMaxCatalogNameLength ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxCatalogNameLength" );
		
		return 0;
	}

	public int getMaxRowSize ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxRowSize" );
		
		return 0;
	}

	public boolean doesMaxRowSizeIncludeBlobs ( ) throws SQLException
	{
		throwUnimplementedException ( "doesMaxRowSizeIncludeBlobs" );
		
		return false;
	}

	public int getMaxStatementLength ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxStatementLength" );
		
		return 0;
	}

	public int getMaxStatements ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxStatements" );
		
		return 0;
	}

	public int getMaxTableNameLength ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxTableNameLength" );
		
		return 0;
	}

	public int getMaxTablesInSelect ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxTablesInSelect" );
		
		return 0;
	}

	public int getMaxUserNameLength ( ) throws SQLException
	{
		throwUnimplementedException ( "getMaxUserNameLength" );
		
		return 0;
	}

	public int getDefaultTransactionIsolation ( ) throws SQLException
	{
		throwUnimplementedException ( "getDefaultTransactionIsolation" );
		
		return 0;
	}

	public boolean supportsTransactions ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsTransactions" );
		
		return false;
	}

	public boolean supportsTransactionIsolationLevel ( int arg0 )
		throws SQLException
	{
		throwUnimplementedException ( "supportsTransactionIsolationLevel" );
		
		return false;
	}

	public boolean supportsDataDefinitionAndDataManipulationTransactions ( )
		throws SQLException
	{
		throwUnimplementedException ( "supportsDataDefinitionAndDataManipulationTransactions" );
		
		return false;
	}

	public boolean supportsDataManipulationTransactionsOnly ( )
		throws SQLException
	{
		throwUnimplementedException ( "supportsDataManipulationTransactionsOnly" );
		
		return false;
	}

	public boolean dataDefinitionCausesTransactionCommit ( )
		throws SQLException
	{
		throwUnimplementedException ( "dataDefinitionCausesTransactionCommit" );
		
		return false;
	}

	public boolean dataDefinitionIgnoredInTransactions ( ) throws SQLException
	{
		throwUnimplementedException ( "dataDefinitionIgnoredInTransactions" );
		
		return false;
	}

	public ResultSet getProcedures ( String arg0, String arg1, String arg2 )
		throws SQLException
	{
		return null;
	}

	public ResultSet getProcedureColumns (
		String arg0,
		String arg1,
		String arg2,
		String arg3)
		throws SQLException
	{
		throwUnimplementedException ( "getProcedureColumns" );
		
		return null;
	}

	/**
	*	All input parameters are ignored and information about ALL available
	*	tables is returned.
	*/
	public ResultSet getTables (
		String strCatalog,
		String strSchemaPattern,
		String strTableNamePattern,
		String [ ] szstrTableTypes )
		throws SQLException
	{
		// This is a fake SQL statement that I will use to generate disconnected resultset
		String			strSQL =	"SELECT " +
								"TABLE_CAT, TABLE_SCHEM, TABLE_NAME, TABLE_TYPE, REMARKS " +
								"FROM JDBC_META_DATA";
		SQLCommand		sqlCommand = null;
		try
		{
			sqlCommand = SQLParser. parse ( strSQL );
		}
		catch ( Exception e )
		{ ; }
		
		// Lets sort table names
		Vector			vctrTableNames = new Vector ( );
		for ( int i = 0; i < m_szsqltMD. length; i++ )
			vctrTableNames. addElement ( m_szsqltMD [ i ]. getDBName ( ) );
			
		Collections. sort ( vctrTableNames ) ;
		
		// Create data for disconnected resultset
		int				nType = opConstants. ALPHANUMERIC;
		opData [ ] [ ]		szopData = new opData [ m_szsqltMD. length ] [ 5 ];
		for ( int i = 0; i < m_szsqltMD. length; i++ )
			szopData [ i ] = new opData [ ] {
									null,
									null,
									new opData ( nType, ( String ) vctrTableNames. elementAt ( i ) ),
									new opData ( nType, "TABLE" ),
									null };

		return new ResultSetImpl ( sqlCommand, szopData );
	}

	public ResultSet getSchemas ( ) throws SQLException
	{
		throwUnimplementedException ( "getSchemas" );
		
		return null;
	}

	public ResultSet getCatalogs ( ) throws SQLException
	{
		// This is a fake SQL statement that I will use to generate disconnected resultset
		String			strSQL =	"SELECT " +
								"TABLE_CAT " +
								"FROM JDBC_META_DATA";
		SQLCommand		sqlCommand = null;
		try
		{
			sqlCommand = SQLParser. parse ( strSQL );
		}
		catch ( Exception e )
		{ ; }
		
		// Create data for disconnected resultset
		int				nTypeAlpha = opConstants. ALPHANUMERIC;
		opData [ ] [ ]		szopData = new opData [ 1 ] [ 1 ];
		szopData [ 0 ] = new opData [ ] { new opData ( nTypeAlpha, "" ) };

		return new ResultSetImpl ( sqlCommand, szopData );
	}

	public ResultSet getTableTypes ( ) throws SQLException
	{
		throwUnimplementedException ( "getTableTypes" );
		
		return null;
	}

	/**
	*	Catalog, schema pattern and column name pattern are completely ignored.
	*	Comlpete, exact table name should be passed as table name pattern and info for all
	*	columns for that table are returned.
	*/
	public ResultSet getColumns (
		String strCatalog,
		String strSchemaPattern,
		String strTableNamePattern,
		String strColumnNamePattern ) throws SQLException
	{
		// This is a fake SQL statement that I will use to generate disconnected resultset
		String			strSQL =	"SELECT " +
								"TABLE_CAT, TABLE_SCHEM, TABLE_NAME, COLUMN_NAME, " +
								"DATA_TYPE, TYPE_NAME, COLUMN_SIZE, BUFFER_LENGTH, " +
								"DECIMAL_DIGITS, NUM_PREC_RADIX, NULLABLE, REMARKS, " +
								"COLUMN_DEF, SQL_DATA_TYPE, SQL_DATETIME_SUB, " +
								"CHAR_OCTET_LENGTH, ORDINAL_POSITION, IS_NULLABLE " +
								"FROM JDBC_META_DATA";
		SQLCommand		sqlCommand = null;
		try
		{
			sqlCommand = SQLParser. parse ( strSQL );
		}
		catch ( Exception e )
		{ ; }
		
		SQLTableMetaData		sqltMD = getTable ( strTableNamePattern );
		if ( sqltMD == null )
			throw new SQLException (
								"com.fourd.jdbc.DatabaseMetaDataImpl::getColumns: " +
								"Table " + strTableNamePattern + " not found" );
								
		// No need to sort by columns' ordinal position because they are already sorted
		SQLColumnMetaData [ ]		szsqlcMD = sqltMD. getColumns ( );

		// Create data for disconnected resultset
		int				nTypeAlpha = opConstants. ALPHANUMERIC;
		int				nTypeInt = opConstants. INTEGER; // java short
		int				nTypeLongInt = opConstants. LONGINTEGER; // java int
		opData [ ] [ ]		szopData = new opData [ szsqlcMD. length ] [ 18 ];
		for ( int i = 0; i < szsqlcMD. length; i++ )
			szopData [ i ] = new opData [ ] {
									null, // TABLE_CAT
									null, // TABLE_SCHEM
									new opData ( nTypeAlpha, strTableNamePattern ), // TABLE_NAME
									new opData ( nTypeAlpha, szsqlcMD [ i ]. getDBName ( ) ), // COLUMN_NAME
									new opData (
											nTypeInt,
											( short ) SQLTools. map4DTypeToSQLType ( szsqlcMD [ i ]. getopType ( ) ) ), // DATA_TYPE
									new opData (
											nTypeAlpha,
											SQLTools. get4DTypeName ( szsqlcMD [ i ]. getopType ( ) ) ), // TYPE_NAME
									new opData ( nTypeLongInt, ( int ) -1 ), // COLUMN_SIZE
									null, // BUFFER_LENGTH
									new opData ( nTypeLongInt, ( int ) -1 ), // DECIMAL_DIGITS
									new opData ( nTypeLongInt, ( int ) 10 ), // NUM_PREC_RADIX
									new opData ( nTypeLongInt, columnNoNulls ), // NULLABLE
									null, // REMARKS
									null, // COLUMN_DEF
									null, // SQL_DATA_TYPE
									null, // SQL_DATETIME_SUB
									new opData ( nTypeLongInt, ( int ) -1 ), // CHAR_OCTET_LENGTH
									new opData ( nTypeLongInt, i + 1 ), // ORDINAL_POSITION
									new opData ( nTypeAlpha, "NO" ) // IS_NULLABLE
									};

		return new ResultSetImpl ( sqlCommand, szopData );
	}

	public ResultSet getColumnPrivileges(
		String arg0,
		String arg1,
		String arg2,
		String arg3)
		throws SQLException
	{
		throwUnimplementedException ( "getColumnPrivileges" );
		
		return null;
	}

	public ResultSet getTablePrivileges(String arg0, String arg1, String arg2)
		throws SQLException
	{
		throwUnimplementedException ( "getTablePrivileges" );
		
		return null;
	}

	public ResultSet getBestRowIdentifier(
		String arg0,
		String arg1,
		String arg2,
		int arg3,
		boolean arg4)
		throws SQLException
	{
		throwUnimplementedException ( "getBestRowIdentifier" );
		
		return null;
	}

	public ResultSet getVersionColumns(String arg0, String arg1, String arg2)
		throws SQLException
	{
		throwUnimplementedException ( "getVersionColumns" );
		
		return null;
	}

	public ResultSet getPrimaryKeys ( String arg0, String arg1, String arg2 )
		throws SQLException
	{
		/*
			TODO: implement properly. Indexed fields should be considered as primary keys.
		*/
		
		// This is a fake SQL statement that I will use to generate disconnected resultset
		String			strSQL ="SELECT " +
								"TABLE_CAT, TABLE_SCHEM, TABLE_NAME, COLUMN_NAME, " +
								"KEY_SEQ, PK_NAME " +
								"FROM JDBC_META_DATA";
		SQLCommand		sqlCommand = null;
		try
		{
			sqlCommand = SQLParser. parse ( strSQL );
		}
		catch ( Exception e )
		{ ; }

		// Create data for disconnected resultset
		opData [ ] [ ]		szopData = new opData [ 1 ] [ 6 ];
		
		return new ResultSetImpl ( sqlCommand, szopData );
	}

	public ResultSet getImportedKeys ( String arg0, String arg1, String arg2 )
		throws SQLException
	{
		throwUnimplementedException ( "getImportedKeys" );
		
		return null;
	}

	public ResultSet getExportedKeys ( String arg0, String arg1, String arg2 )
		throws SQLException
	{
		throwUnimplementedException ( "getExportedKeys" );
		
		return null;
	}

	public ResultSet getCrossReference(
		String arg0,
		String arg1,
		String arg2,
		String arg3,
		String arg4,
		String arg5)
		throws SQLException
	{
		throwUnimplementedException ( "getCrossReference" );
		
		return null;
	}

	public ResultSet getTypeInfo ( ) throws SQLException
	{
		// This is a fake SQL statement that I will use to generate disconnected resultset
		String			strSQL ="SELECT " +
								"TYPE_NAME, DATA_TYPE, PRECISION, LITERAL_PREFIX, " +
								"LITERAL_SUFFIX, CREATE_PARAMS, NULLABLE, CASE_SENSITIVE, " +
								"SEARCHABLE, UNSIGNED_ATTRIBUTE, FIXED_PREC_SCALE, AUTO_INCREMENT, " +
								"LOCAL_TYPE_NAME, MINIMUM_SCALE, MAXIMUM_SCALE, SQL_DATA_TYPE, " +
								"SQL_DATETIME_SUB, NUM_PREC_RADIX " +
								"FROM JDBC_META_DATA";
		SQLCommand		sqlCommand = null;
		try
		{
			sqlCommand = SQLParser. parse ( strSQL );
		}
		catch ( Exception e )
		{ ; }

		// Create data for disconnected resultset
		int					nTypeAlpha = opConstants. ALPHANUMERIC;
		int					nTypeInt = opConstants. INTEGER; // java short
		int					nTypeLongInt = opConstants. LONGINTEGER; // java int
		int					nTypeBoolean = opConstants. BOOLEAN;
		opData [ ] [ ]		szopData = new opData [ 9 ] [ 18 ];
		
		szopData [ 0 ] = new opData [ ] {
									new opData ( nTypeAlpha, "BOOLEAN" ), // TYPE_NAME
									new opData ( nTypeInt, ( short ) java.sql.Types.BIT ), // DATA_TYPE
									new opData ( nTypeLongInt, 10 ), // PRECISION - WHATEVER
									null, // LITERAL_PREFIX
									null, // LITERAL_SUFFIX
									null, // CREATE_PARAMS
									new opData ( nTypeLongInt, typeNullableUnknown ), // NULLABLE
									new opData ( nTypeBoolean, false ), // CASE_SENSITIVE
									new opData ( nTypeInt, typePredBasic ), // SEARCHABLE
									new opData ( nTypeBoolean, true ), // UNSIGNED_ATTRIBUTE
									new opData ( nTypeBoolean, false ), // FIXED_PREC_SCALE
									new opData ( nTypeBoolean, false ), // AUTO_INCREMENT
									new opData ( nTypeAlpha, "BOOLEAN" ), // LOCAL_TYPE_NAME
									new opData ( nTypeInt, ( short ) 1 ), // MINIMUM_SCALE
									new opData ( nTypeInt, ( short ) 10 ), // MAXIMUM_SCALE
									new opData ( nTypeLongInt, 10 ), // SQL_DATA_TYPE - WHATEVER
									new opData ( nTypeLongInt, 10 ), // SQL_DATETIME_SUB - WHATEVER
									new opData ( nTypeLongInt, 10 ) }; // NUM_PREC_RADIX - WHATEVER

		szopData [ 1 ] = new opData [ ] {
									new opData ( nTypeAlpha, "TEXT" ), // TYPE_NAME
									new opData ( nTypeInt, ( short ) java.sql.Types.LONGVARCHAR ), // DATA_TYPE
									new opData ( nTypeLongInt, 10 ), // PRECISION - WHATEVER
									new opData ( nTypeAlpha, "'" ), // LITERAL_PREFIX
									new opData ( nTypeAlpha, "'" ), // LITERAL_SUFFIX
									null, // CREATE_PARAMS
									new opData ( nTypeLongInt, typeNullableUnknown ), // NULLABLE
									new opData ( nTypeBoolean, false ), // CASE_SENSITIVE
									new opData ( nTypeInt, typePredBasic ), // SEARCHABLE
									new opData ( nTypeBoolean, true ), // UNSIGNED_ATTRIBUTE
									new opData ( nTypeBoolean, false ), // FIXED_PREC_SCALE
									new opData ( nTypeBoolean, false ), // AUTO_INCREMENT
									new opData ( nTypeAlpha, "TEXT" ), // LOCAL_TYPE_NAME
									new opData ( nTypeInt, ( short ) 1 ), // MINIMUM_SCALE
									new opData ( nTypeInt, ( short ) 10 ), // MAXIMUM_SCALE
									new opData ( nTypeLongInt, 10 ), // SQL_DATA_TYPE - WHATEVER
									new opData ( nTypeLongInt, 10 ), // SQL_DATETIME_SUB - WHATEVER
									new opData ( nTypeLongInt, 10 ) }; // NUM_PREC_RADIX - WHATEVER

		szopData [ 2 ] = new opData [ ] {
									new opData ( nTypeAlpha, "LONGINTEGER" ), // TYPE_NAME
									new opData ( nTypeInt, ( short ) java.sql.Types.INTEGER ), // DATA_TYPE
									new opData ( nTypeLongInt, 10 ), // PRECISION - WHATEVER
									null, // LITERAL_PREFIX
									null, // LITERAL_SUFFIX
									null, // CREATE_PARAMS
									new opData ( nTypeLongInt, typeNullableUnknown ), // NULLABLE
									new opData ( nTypeBoolean, false ), // CASE_SENSITIVE
									new opData ( nTypeInt, typePredBasic ), // SEARCHABLE
									new opData ( nTypeBoolean, false ), // UNSIGNED_ATTRIBUTE
									new opData ( nTypeBoolean, false ), // FIXED_PREC_SCALE
									new opData ( nTypeBoolean, true ), // AUTO_INCREMENT
									new opData ( nTypeAlpha, "LONGINTEGER" ), // LOCAL_TYPE_NAME
									new opData ( nTypeInt, ( short ) 1 ), // MINIMUM_SCALE
									new opData ( nTypeInt, ( short ) 10 ), // MAXIMUM_SCALE
									new opData ( nTypeLongInt, 10 ), // SQL_DATA_TYPE - WHATEVER
									new opData ( nTypeLongInt, 10 ), // SQL_DATETIME_SUB - WHATEVER
									new opData ( nTypeLongInt, 10 ) }; // NUM_PREC_RADIX - WHATEVER

		szopData [ 3 ] = new opData [ ] {
									new opData ( nTypeAlpha, "INTEGER" ), // TYPE_NAME
									new opData ( nTypeInt, ( short ) java.sql.Types.SMALLINT ), // DATA_TYPE
									new opData ( nTypeLongInt, 10 ), // PRECISION - WHATEVER
									null, // LITERAL_PREFIX
									null, // LITERAL_SUFFIX
									null, // CREATE_PARAMS
									new opData ( nTypeLongInt, typeNullableUnknown ), // NULLABLE
									new opData ( nTypeBoolean, false ), // CASE_SENSITIVE
									new opData ( nTypeInt, typePredBasic ), // SEARCHABLE
									new opData ( nTypeBoolean, false ), // UNSIGNED_ATTRIBUTE
									new opData ( nTypeBoolean, false ), // FIXED_PREC_SCALE
									new opData ( nTypeBoolean, true ), // AUTO_INCREMENT
									new opData ( nTypeAlpha, "INTEGER" ), // LOCAL_TYPE_NAME
									new opData ( nTypeInt, ( short ) 1 ), // MINIMUM_SCALE
									new opData ( nTypeInt, ( short ) 10 ), // MAXIMUM_SCALE
									new opData ( nTypeLongInt, 10 ), // SQL_DATA_TYPE - WHATEVER
									new opData ( nTypeLongInt, 10 ), // SQL_DATETIME_SUB - WHATEVER
									new opData ( nTypeLongInt, 10 ) }; // NUM_PREC_RADIX - WHATEVER

		szopData [ 4 ] = new opData [ ] {
									new opData ( nTypeAlpha, "REAL" ), // TYPE_NAME
									new opData ( nTypeInt, ( short ) java.sql.Types.DOUBLE ), // DATA_TYPE
									new opData ( nTypeLongInt, 10 ), // PRECISION - WHATEVER
									null, // LITERAL_PREFIX
									null, // LITERAL_SUFFIX
									null, // CREATE_PARAMS
									new opData ( nTypeLongInt, typeNullableUnknown ), // NULLABLE
									new opData ( nTypeBoolean, false ), // CASE_SENSITIVE
									new opData ( nTypeInt, typePredBasic ), // SEARCHABLE
									new opData ( nTypeBoolean, false ), // UNSIGNED_ATTRIBUTE
									new opData ( nTypeBoolean, false ), // FIXED_PREC_SCALE
									new opData ( nTypeBoolean, true ), // AUTO_INCREMENT
									new opData ( nTypeAlpha, "REAL" ), // LOCAL_TYPE_NAME
									new opData ( nTypeInt, ( short ) 1 ), // MINIMUM_SCALE
									new opData ( nTypeInt, ( short ) 10 ), // MAXIMUM_SCALE
									new opData ( nTypeLongInt, 10 ), // SQL_DATA_TYPE - WHATEVER
									new opData ( nTypeLongInt, 10 ), // SQL_DATETIME_SUB - WHATEVER
									new opData ( nTypeLongInt, 10 ) }; // NUM_PREC_RADIX - WHATEVER

		szopData [ 5 ] = new opData [ ] {
									new opData ( nTypeAlpha, "ALPHANUMERIC" ), // TYPE_NAME
									new opData ( nTypeInt, ( short ) java.sql.Types.VARCHAR ), // DATA_TYPE
									new opData ( nTypeLongInt, 10 ), // PRECISION - WHATEVER
									new opData ( nTypeAlpha, "'" ), // LITERAL_PREFIX
									new opData ( nTypeAlpha, "'" ), // LITERAL_SUFFIX
									null, // CREATE_PARAMS
									new opData ( nTypeLongInt, typeNullableUnknown ), // NULLABLE
									new opData ( nTypeBoolean, false ), // CASE_SENSITIVE
									new opData ( nTypeInt, typePredBasic ), // SEARCHABLE
									new opData ( nTypeBoolean, true ), // UNSIGNED_ATTRIBUTE
									new opData ( nTypeBoolean, false ), // FIXED_PREC_SCALE
									new opData ( nTypeBoolean, false ), // AUTO_INCREMENT
									new opData ( nTypeAlpha, "ALPHANUMERIC" ), // LOCAL_TYPE_NAME
									new opData ( nTypeInt, ( short ) 1 ), // MINIMUM_SCALE
									new opData ( nTypeInt, ( short ) 10 ), // MAXIMUM_SCALE
									new opData ( nTypeLongInt, 10 ), // SQL_DATA_TYPE - WHATEVER
									new opData ( nTypeLongInt, 10 ), // SQL_DATETIME_SUB - WHATEVER
									new opData ( nTypeLongInt, 10 ) }; // NUM_PREC_RADIX - WHATEVER

		szopData [ 6 ] = new opData [ ] {
									new opData ( nTypeAlpha, "DATE" ), // TYPE_NAME
									new opData ( nTypeInt, ( short ) java.sql.Types.DATE ), // DATA_TYPE
									new opData ( nTypeLongInt, 10 ), // PRECISION - WHATEVER
									new opData ( nTypeAlpha, "'" ), // LITERAL_PREFIX
									new opData ( nTypeAlpha, "'" ), // LITERAL_SUFFIX
									null, // CREATE_PARAMS
									new opData ( nTypeLongInt, typeNullableUnknown ), // NULLABLE
									new opData ( nTypeBoolean, false ), // CASE_SENSITIVE
									new opData ( nTypeInt, typePredBasic ), // SEARCHABLE
									new opData ( nTypeBoolean, true ), // UNSIGNED_ATTRIBUTE
									new opData ( nTypeBoolean, false ), // FIXED_PREC_SCALE
									new opData ( nTypeBoolean, false ), // AUTO_INCREMENT
									new opData ( nTypeAlpha, "DATE" ), // LOCAL_TYPE_NAME
									new opData ( nTypeInt, ( short ) 1 ), // MINIMUM_SCALE
									new opData ( nTypeInt, ( short ) 10 ), // MAXIMUM_SCALE
									new opData ( nTypeLongInt, 10 ), // SQL_DATA_TYPE - WHATEVER
									new opData ( nTypeLongInt, 10 ), // SQL_DATETIME_SUB - WHATEVER
									new opData ( nTypeLongInt, 10 ) }; // NUM_PREC_RADIX - WHATEVER

		szopData [ 7 ] = new opData [ ] {
									new opData ( nTypeAlpha, "TIME" ), // TYPE_NAME
									new opData ( nTypeInt, ( short ) java.sql.Types.TIME ), // DATA_TYPE
									new opData ( nTypeLongInt, 10 ), // PRECISION - WHATEVER
									new opData ( nTypeAlpha, "'" ), // LITERAL_PREFIX
									new opData ( nTypeAlpha, "'" ), // LITERAL_SUFFIX
									null, // CREATE_PARAMS
									new opData ( nTypeLongInt, typeNullableUnknown ), // NULLABLE
									new opData ( nTypeBoolean, false ), // CASE_SENSITIVE
									new opData ( nTypeInt, typePredBasic ), // SEARCHABLE
									new opData ( nTypeBoolean, true ), // UNSIGNED_ATTRIBUTE
									new opData ( nTypeBoolean, false ), // FIXED_PREC_SCALE
									new opData ( nTypeBoolean, false ), // AUTO_INCREMENT
									new opData ( nTypeAlpha, "TIME" ), // LOCAL_TYPE_NAME
									new opData ( nTypeInt, ( short ) 1 ), // MINIMUM_SCALE
									new opData ( nTypeInt, ( short ) 10 ), // MAXIMUM_SCALE
									new opData ( nTypeLongInt, 10 ), // SQL_DATA_TYPE - WHATEVER
									new opData ( nTypeLongInt, 10 ), // SQL_DATETIME_SUB - WHATEVER
									new opData ( nTypeLongInt, 10 ) }; // NUM_PREC_RADIX - WHATEVER

		szopData [ 8 ] = new opData [ ] {
									new opData ( nTypeAlpha, "BLOB" ), // TYPE_NAME
									new opData ( nTypeInt, ( short ) java.sql.Types.BLOB ), // DATA_TYPE
									new opData ( nTypeLongInt, 10 ), // PRECISION - WHATEVER
									null, // LITERAL_PREFIX
									null, // LITERAL_SUFFIX
									null, // CREATE_PARAMS
									new opData ( nTypeLongInt, typeNullableUnknown ), // NULLABLE
									new opData ( nTypeBoolean, false ), // CASE_SENSITIVE
									new opData ( nTypeInt, typePredBasic ), // SEARCHABLE
									new opData ( nTypeBoolean, true ), // UNSIGNED_ATTRIBUTE
									new opData ( nTypeBoolean, false ), // FIXED_PREC_SCALE
									new opData ( nTypeBoolean, false ), // AUTO_INCREMENT
									new opData ( nTypeAlpha, "BLOB" ), // LOCAL_TYPE_NAME
									new opData ( nTypeInt, ( short ) 1 ), // MINIMUM_SCALE
									new opData ( nTypeInt, ( short ) 10 ), // MAXIMUM_SCALE
									new opData ( nTypeLongInt, 10 ), // SQL_DATA_TYPE - WHATEVER
									new opData ( nTypeLongInt, 10 ), // SQL_DATETIME_SUB - WHATEVER
									new opData ( nTypeLongInt, 10 ) }; // NUM_PREC_RADIX - WHATEVER

		return new ResultSetImpl ( sqlCommand, szopData );
	}

	public ResultSet getIndexInfo(
		String arg0,
		String arg1,
		String arg2,
		boolean arg3,
		boolean arg4)
		throws SQLException
	{
		throwUnimplementedException ( "getIndexInfo" );
		
		return null;
	}

	public boolean supportsResultSetType ( int arg0 ) throws SQLException
	{
		throwUnimplementedException ( "supportsResultSetType" );
		
		return false;
	}

	public boolean supportsResultSetConcurrency(int arg0, int arg1)
		throws SQLException
	{
		// TODO: implemet completely
		
		return true;
	}

	public boolean ownUpdatesAreVisible ( int arg0 ) throws SQLException
	{
		throwUnimplementedException ( "ownUpdatesAreVisible" );
		
		return false;
	}

	public boolean ownDeletesAreVisible ( int arg0 ) throws SQLException
	{
		throwUnimplementedException ( "ownDeletesAreVisible" );
		
		return false;
	}

	public boolean ownInsertsAreVisible ( int arg0 ) throws SQLException
	{
		throwUnimplementedException ( "ownInsertsAreVisible" );
		
		return false;
	}

	public boolean othersUpdatesAreVisible ( int arg0 ) throws SQLException
	{
		throwUnimplementedException ( "othersUpdatesAreVisible" );
		
		return false;
	}

	public boolean othersDeletesAreVisible ( int arg0 ) throws SQLException
	{
		throwUnimplementedException ( "othersDeletesAreVisible" );
		
		return false;
	}

	public boolean othersInsertsAreVisible ( int arg0 ) throws SQLException
	{
		throwUnimplementedException ( "othersInsertsAreVisible" );
		
		return false;
	}

	public boolean updatesAreDetected ( int arg0 ) throws SQLException
	{
		throwUnimplementedException ( "updatesAreDetected" );
		
		return false;
	}

	public boolean deletesAreDetected ( int arg0 ) throws SQLException
	{
		throwUnimplementedException ( "deletesAreDetected" );
		
		return false;
	}

	public boolean insertsAreDetected ( int arg0 ) throws SQLException
	{
		throwUnimplementedException ( "insertsAreDetected" );
		
		return false;
	}

	public boolean supportsBatchUpdates ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsBatchUpdates" );
		
		return false;
	}

	public ResultSet getUDTs ( String arg0, String arg1, String arg2, int[] arg3 )
		throws SQLException
	{
		throwUnimplementedException ( "getUDTs" );
		
		return null;
	}

	public Connection getConnection ( ) throws SQLException
	{
		throwUnimplementedException ( "getConnection" );
		
		return null;
	}

	public boolean supportsSavepoints ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsSavepoints" );
		
		return false;
	}

	public boolean supportsNamedParameters ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsNamedParameters" );
		
		return false;
	}

	public boolean supportsMultipleOpenResults ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsMultipleOpenResults" );
		
		return false;
	}

	public boolean supportsGetGeneratedKeys ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsGetGeneratedKeys" );
		
		return false;
	}

	public ResultSet getSuperTypes ( String arg0, String arg1, String arg2 )
		throws SQLException
	{
		throwUnimplementedException ( "getSuperTypes" );
		
		return null;
	}

	public ResultSet getSuperTables ( String arg0, String arg1, String arg2 )
		throws SQLException
	{
		throwUnimplementedException ( "getSuperTables" );
		
		return null;
	}

	public ResultSet getAttributes (
		String arg0,
		String arg1,
		String arg2,
		String arg3)
		throws SQLException
	{
		throwUnimplementedException ( "getAttributes" );
		
		return null;
	}

	public boolean supportsResultSetHoldability ( int arg0 ) throws SQLException
	{
		throwUnimplementedException ( "supportsResultSetHoldability" );
		
		return false;
	}

	public int getResultSetHoldability ( ) throws SQLException
	{
		throwUnimplementedException ( "getResultSetHoldability" );
		
		return 0;
	}

	public int getDatabaseMajorVersion ( ) throws SQLException
	{
		throwUnimplementedException ( "getDatabaseMajorVersion" );
		
		return 0;
	}

	public int getDatabaseMinorVersion ( ) throws SQLException
	{
		throwUnimplementedException ( "getDatabaseMinorVersion" );
		
		return 0;
	}

	public int getJDBCMajorVersion ( ) throws SQLException
	{
		throwUnimplementedException ( "getJDBCMajorVersion" );
		
		return 0;
	}

	public int getJDBCMinorVersion ( ) throws SQLException
	{
		throwUnimplementedException ( "getJDBCMinorVersion" );
		
		return 0;
	}

	public int getSQLStateType ( ) throws SQLException
	{
		throwUnimplementedException ( "getSQLStateType" );
		
		return 0;
	}

	public boolean locatorsUpdateCopy ( ) throws SQLException
	{
		throwUnimplementedException ( "locatorsUpdateCopy" );
		
		return false;
	}

	public boolean supportsStatementPooling ( ) throws SQLException
	{
		throwUnimplementedException ( "supportsStatementPooling" );
		
		return false;
	}
	
	private void s_e_p_a_r_a_t_o_r__________________________1 ( )
	{
		;
	}
	
	private void loadMetaData ( ) throws SQLException
	{
		m_szsqltMD = loadTablesMetaData ( );
	}
	
	private SQLTableMetaData [ ] loadTablesMetaData ( ) throws SQLException
	{
		SQLTableMetaData [ ]			szsqltMD = null;
		opTableArray					optaAllTables = null;
		try
		{
			optaAllTables = new opTableArray ( m_opProcess. CountTables ( ) );
			m_opProcess. GetTableNames ( optaAllTables );
		}
		catch ( opException opE )
		{
			throw new SQLException ( "com.fourd.jdbc.DatabaseMetaDataImpl::loadTablesMetaData: " + opE. GetErrorMsg ( ) );
		}

		szsqltMD = new SQLTableMetaData [ optaAllTables. mSize ];
		for ( int i = 0; i < szsqltMD. length; i++ )
			szsqltMD [ i ] = loadTableMetaData ( optaAllTables. mTableArray [ i ] );
			
		return szsqltMD;
	}
	
	private SQLTableMetaData loadTableMetaData ( opTable opT ) throws SQLException
	{
		SQLTableMetaData			sqltMD = new SQLTableMetaData ( );
		try
		{
			m_opProcess. GetTableProperties ( opT );
			sqltMD. setNumber ( opT. mTableNumber );
			sqltMD. setDBName ( opT. mTableName );
		}
		catch ( opException opE )
		{
			throw new SQLException ( "com.fourd.jdbc.DatabaseMetaDataImpl::loadTableMetaData: " + opE. GetErrorMsg ( ) );
		}
		
		sqltMD. addColumns ( loadColumnsMetaData ( opT ) );
		
		return sqltMD;
	}
	
	private SQLColumnMetaData [ ] loadColumnsMetaData ( opTable opT )
	{
		opField [ ]				szopFields = opT. mFieldList. mFieldArray;
		SQLColumnMetaData [ ]		szsqlcMD = new SQLColumnMetaData [ szopFields. length ];
// System. out. println ( "Column info - START" );
		for ( int i = 0; i < szopFields. length; i++ )
		{
			szsqlcMD [ i ] = new SQLColumnMetaData ( szopFields [ i ]. mFieldName );
			szsqlcMD [ i ]. setNumber ( szopFields [ i ]. mFieldNumber );
			szsqlcMD [ i ]. setopType ( szopFields [ i ]. mFieldType );
// System. out. println ( "Column #" + i + " is of type " + szopFields [ i ]. mFieldType );
		}
// System. out. println ( "Column info - END" );
			
		return szsqlcMD;
	}
	
	/**
	*	Loads table names and numbers but does not load column information.
	*/
	private void lazyLoadMetaData ( ) throws SQLException
	{
		m_szsqltMD = lazyLoadTablesMetaData ( );
	}
	
	private SQLTableMetaData [ ] lazyLoadTablesMetaData ( ) throws SQLException
	{
		SQLTableMetaData [ ]			szsqltMD = null;
		opTableArray					optaAllTables = null;
		try
		{
			optaAllTables = new opTableArray ( m_opProcess. CountTables ( ) );
			m_opProcess. GetTableNames ( optaAllTables );
		}
		catch ( opException opE )
		{
			throw new SQLException ( "com.fourd.jdbc.DatabaseMetaDataImpl::loadTablesMetaData: " + opE. GetErrorMsg ( ) );
		}

		szsqltMD = new SQLTableMetaData [ optaAllTables. mSize ];
		for ( int i = 0; i < szsqltMD. length; i++ )
			szsqltMD [ i ] = lazyLoadTableMetaData ( optaAllTables. mTableArray [ i ] );
			
		return szsqltMD;
	}
	
	private SQLTableMetaData lazyLoadTableMetaData ( opTable opT )
	{
		SQLTableMetaData			sqltMD = new SQLTableMetaData ( this, opT );
// System. out. println ( "opT. mTableNumber == " + opT. mTableNumber );
		sqltMD. setNumber ( opT. mTableNumber );
// System. out. println ( "opT. mTableName == " + opT. mTableName );
		sqltMD. setDBName ( opT. mTableName );
		
		return sqltMD;
	}
	
	/**
	*	This is actually not lazy but real loading. It is called lazy just because
	*	it is used for lazy loaded tables.
	*/
	SQLColumnMetaData [ ] lazyLoadColumnsMetaData ( opTable opT ) throws SQLException
	{
		// First of all lets fetch column info
		try
		{
			synchronized ( m_objDataAccessLock )
			{
				m_opProcess. GetTableProperties ( opT );
			}
		}
		catch ( opException opE )
		{
			throw new SQLException (
								"com.fourd.jdbc.DatabaseMetaDataImpl::lazyLoadColumnsMetaData: " + 
								opE. GetErrorMsg ( ) );
		}
		
		opField [ ]					szopFields = opT. mFieldList. mFieldArray;
		SQLColumnMetaData [ ]		szsqlcMD = new SQLColumnMetaData [ szopFields. length ];
// System. out. println ( "Column info - START" );
		for ( int i = 0; i < szopFields. length; i++ )
		{
			szsqlcMD [ i ] = new SQLColumnMetaData ( szopFields [ i ]. mFieldName );
			szsqlcMD [ i ]. setNumber ( szopFields [ i ]. mFieldNumber );
			szsqlcMD [ i ]. setopType ( szopFields [ i ]. mFieldType );
// System. out. println ( "Column #" + i + " is of type " + szopFields [ i ]. mFieldType );
		}
// System. out. println ( "Column info - END" );
			
		return szsqlcMD;
	}
	
	/**
	*	Determines presence of a table with a given name in the database.
	*/
	boolean tableExists ( String strTableName )
	{
		if ( strTableName == null )
			return false;
		
		SQLTableMetaData		sqltMD = null;	
		for ( int i = 0; i < m_szsqltMD. length; i++ )
		{
			if ( m_szsqltMD [ i ]. getDBName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) )
				return true;
		}
		
		return false;
	}
	
	boolean columnExists ( String strTableName, String strColumnName ) throws SQLException
	{
		if ( strTableName == null )
			throw new SQLException ( "com.fourd.jdbc.DatabaseMetaDataImpl::columnExists: Table name should not be null" );
			
		SQLTableMetaData		sqltMD = getTable ( strTableName );
		if ( sqltMD == null )
			throw new SQLException (
								"com.fourd.jdbc.DatabaseMetaDataImpl::columnExists: Table " +
								 strTableName + " does not exist" );
			
		return sqltMD. columnExists ( strColumnName );
	}
	
	SQLTableMetaData getTable ( String strTableName ) throws SQLException
	{
		if ( strTableName == null )
			throw new SQLException ( "com.fourd.jdbc.DatabaseMetaDataImpl::columnExists: Table name should not be null" );
			
		for ( int i = 0; i < m_szsqltMD. length; i++ )
			if ( m_szsqltMD [ i ]. getDBName ( ). toUpperCase ( ). equals ( strTableName. toUpperCase ( ) ) )
				return m_szsqltMD [ i ];
			
		return null;
	}
	
	private void throwUnimplementedException ( String strMethodName ) throws SQLException
	{
		throw new SQLException (
							"com.fourd.jdbc.DatabaseMetaDataImpl: " +
							"Method " + strMethodName + " is not implemented" );
	}

	private void throwUnsupportedException ( ) throws SQLException
	{
		throw new SQLException (
							"com.fourd.jdbc.DatabaseMetaDataImpl: " +
							"Method/Feature is not supported" );
	}
}