/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.sql.*;
import java.util.*;

import j4d.open.*;

public class DriverImpl implements Driver
{
	static final int				MAJOR_VERSION = 1;
	static final int				MINOR_VERSION = 0;
	static final String				DRIVER_NAME = "4D JDBC Driver v1.0";
	static final String				VERSION =	String. valueOf ( MAJOR_VERSION ) +
												"." +
												String. valueOf ( MINOR_VERSION );
	
static
{
	try
	{
		java.sql.DriverManager.registerDriver ( new DriverImpl ( ) );
	}
	catch ( SQLException sqlE )
	{
		System. err. println ( "Unable to register driver :" );
		System. err. println ( "\t" + sqlE. getMessage ( ) );
	}
}

	/**
	*	Returns true if the driver thinks that it can open a connection  to the given URL.
	*
	*	Proper format would be:
	*		jdbc:4d:<ip_or_dnsName>[:<port_number>]
	*/
	public boolean acceptsURL ( String strURL )
	{
		return	strURL != null &&
				strURL. startsWith ( "jdbc:4d:" ) &&
				strURL. length ( ) > 8;
	}

	/**
	*	Attempts to make a database connection to the given URL.
	*
	*	Proper format for strURL is:
	*		jdbc:4d:<ip_or_dnsName>[:<port_number>]
	*/
	public Connection connect ( String strURL, Properties propsInfo ) throws SQLException
	{
		if ( !acceptsURL ( strURL ) )
			throw new SQLException ( "com.fourd.jdbc.DriverImpl::connect: Unrecognizable URL - " + strURL );
			
		if ( propsInfo == null )
			throw new SQLException ( "com.fourd.jdbc.DriverImpl::connect: Connection properties can not be null" );
			
		// Lets remove jdbc:4d: prefix
		strURL = strURL. substring ( 8 );
		
		ConnectionImpl		connImpl = null;
		try
		{
			connImpl = new ConnectionImpl ( strURL, propsInfo );
		}
		catch ( Exception e )
		{
			throw new SQLException ( e. getMessage ( ) );
		}
		
		return connImpl;
	}

	/**
	*	Gets the driver's major version number.
	*/
	public int getMajorVersion ( )
	{
		return MAJOR_VERSION;
	}

	/**
	*	Gets the driver's minor version number.
	*/
	public int getMinorVersion ( )
	{
		return MINOR_VERSION;
	}

	/**
	*	Gets information about the possible properties for this driver.
	*/
	public DriverPropertyInfo [ ] getPropertyInfo ( String strURL, Properties propsInfo ) throws SQLException
	{
		throwUnimplementedException ( );
		
		return null;
	}

	/**
	*	Reports whether this driver is a genuine JDBC  COMPLIANT TM driver.
	*/
	public boolean jdbcCompliant ( )
	{
		return false;
	}

	private void throwUnimplementedException ( ) throws SQLException
	{
		throw new SQLException ( "com.fourd.jdbc.ConnectionImpl: Method is not implemented" );
	}
}