/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.io.*;
import java.math.*;
import java.net.*;
import java.sql.*;
import java.util.*;
import j4d.open.*;

/**
*	Note: prepared statements can be optimized by creating a named selection the first time
*	execute is called WITHOUT applying WHERE part. Parameters (?) can be only in WHERE
*	part so when the execute method is called once again, all I have to do is to 
*	restore named selection and apply WHERE part. Oh, but that is not going to work 
*	is the result set from the first execute was changed with delete or create new row...
*/
class PreparedStatementImpl extends StatementImpl implements PreparedStatement
{
	protected String				m_strSQL = null;
	protected SQLCommand			m_sqlCommand = null;
	
	PreparedStatementImpl (
							ConnectionImpl connImpl,
							opConnection opC,
							opProcess opP,
							DatabaseMetaDataImpl dbMD,
							Object objDataAccessLock,
							String strSQL ) throws SQLException
	{
		super (	connImpl,
				opC,
				opP,
				dbMD,
				objDataAccessLock );

		m_strSQL = strSQL;
		try
		{
			m_sqlCommand = SQLParser. parse ( m_strSQL );
		}
		catch ( Exception e )
		{
			e. printStackTrace ( );
			
			throw new SQLException ( e. getMessage ( ) );
		}
	}
	
	public void addBatch ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "addBatch" );
	}
	
	public void clearParameters ( ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "clearParameters" );
	}
	
	public boolean execute ( ) throws SQLException
	{
		validateOpened ( );
		
		if ( m_sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_SELECT )
			executeSELECTCommand ( m_sqlCommand );
		else if ( m_sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_DELETE )
			executeDELETECommand ( m_sqlCommand );
		else if ( m_sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_INSERT )
			executeINSERTCommand ( m_sqlCommand );
		else if ( m_sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_UPDATE )
			executeUPDATECommand ( m_sqlCommand );
		else
			throw new SQLException (
								"com.fourd.jdbc.PreparedStatementImpl::execute: " +
								"Unsupported query" );
		
		// TODO: implement properly and make sure that return value is correct
		return false;
	}

	public ResultSet executeQuery ( ) throws SQLException
	{
		validateOpened ( );
		
		if ( m_sqlCommand. getType ( ) != SQLCommand. CMD_TYPE_SELECT )
			// execute Query should only be able to perform SELECT queries in the current implementation
			throw new SQLException ( "com.fourd.jdbc.PreparedStatementImpl::executeQuery: Unsupported query" );

		return executeSELECTCommand ( m_sqlCommand );
	}

	public int executeUpdate ( ) throws SQLException
	{
		validateOpened ( );
		
		if ( m_sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_DELETE )
			return executeDELETECommand ( m_sqlCommand );
		else if ( m_sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_INSERT )
			return executeINSERTCommand ( m_sqlCommand );
		else if ( m_sqlCommand. getType ( ) == SQLCommand. CMD_TYPE_UPDATE )
			return executeUPDATECommand ( m_sqlCommand );
		else
			throw new SQLException (
								"com.fourd.jdbc.PreparedStatementImpl::executeUpdate: " +
								"Unsupported query" );
	}

	public ResultSetMetaData getMetaData ( ) throws SQLException
	{
		validateOpened ( );
		
		return null;
	}

	public void setArray ( int i, Array x ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setArray" );
	}

	public void setAsciiStream ( int nIndex, InputStream  x, int length ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setAsciiStream" );
	}

	public void setBigDecimal ( int nIndex, BigDecimal  x) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setBigDecimal" );
	}

	public void setBinaryStream ( int nIndex, InputStream  x, int length) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setBinaryStream" );
	}

	public void setBlob ( int i, Blob  x) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setBlob" );
	}

	public void setBoolean ( int nIndex, boolean bValue ) throws SQLException
	{
		validateOpened ( );
			
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												new Boolean ( bValue ),
												opConstants. BOOLEAN ) );
	}

	public void setByte ( int nIndex, byte btValue ) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setByte" );
	}

	public void setBytes ( int nIndex, byte [ ] szbtValue ) throws SQLException
	{
		validateOpened ( );
			
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												szbtValue,
												opConstants. BLOB ) );
	}

	public void setCharacterStream ( int nIndex, Reader reader, int nLength ) throws SQLException
	{
		validateOpened ( );
		
		char [ ]		szch = new char [ 1 ];
		int				nRead = 0;
		StringBuffer	strbText = new StringBuffer ( "" );
		try
		{
			while ( nRead < nLength && ( reader. read ( szch ) ) != -1 )
			{
				nRead++;
				strbText. append ( szch [ 0 ] );
			}
		}
		catch ( IOException ioE )
		{
			throw new SQLException (
								"PreparedStatementImpl::setCharacterStream: " +
								ioE. getMessage ( ) );
		}
		
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												strbText. toString ( ),
												opConstants. TEXT ) );
	}

	public void setClob ( int i, Clob  x) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setClob" );
	}

	public void setDate ( int nIndex, java.sql.Date dtValue ) throws SQLException
	{
		validateOpened ( );
			
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												dtValue,
												opConstants. DATE ) );
	}

	public void setDate ( int nIndex, java.sql.Date  x, Calendar  cal) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setDate" );
	}

	public void setDouble ( int nIndex, double nValue ) throws SQLException
	{
		validateOpened ( );
			
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												new Double ( nValue ),
												opConstants. REAL ) );
	}

	public void setFloat ( int nIndex, float nValue ) throws SQLException
	{
		validateOpened ( );
		
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												new Float ( nValue ),
												opConstants. REAL ) );
	}

	public void setInt ( int nIndex, int nValue ) throws SQLException
	{
		validateOpened ( );
			
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												new Integer ( nValue ),
												opConstants. LONGINTEGER ) );
	}

	public void setLong ( int nIndex, long x) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setLong" );
	}

	public void setNull ( int nIndex, int sqlType) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setNull" );
	}

	public void setNull ( int paramIndex, int sqlType, String typeName) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setNull" );
	}

	public void setObject ( int nIndex, Object obj ) throws SQLException
	{
		validateOpened ( );
			
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												obj,
												SQLTools. mapJavaClassTo4DType ( obj ) ) );
	}

	public void setObject ( int nIndex, Object obj, int nTargetSqlType ) throws SQLException
	{
		validateOpened ( );
			
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												obj,
												SQLTools. mapSQLTypeTo4DType ( nTargetSqlType ) ) );
	}

	public void setObject ( int nIndex, Object obj, int nTargetSqlType, int nScale ) throws SQLException
	{
		validateOpened ( );
		
		// TODO: should take care of nScale when necessary
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												obj,
												SQLTools. mapSQLTypeTo4DType ( nTargetSqlType ) ) );
	}

	public void setRef ( int i, Ref  x) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setRef" );
	}

	public void setShort ( int nIndex, short nValue ) throws SQLException
	{
		validateOpened ( );
			
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												new Short ( nValue ),
												opConstants. INTEGER ) );
	}

	public void setString ( int nIndex, String strValue ) throws SQLException
	{
		validateOpened ( );
		
		/* TODO : Currently there is no way to distinguish here between 4D alphanumeric and 
		*	4D text. I should really check the column's type and set type properly in the 
		*	StatementImpl. executeXXX methods.
		*/
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												strValue,
												opConstants. ALPHANUMERIC ) );
	}

	public void setTime ( int nIndex, Time tmValue ) throws SQLException
	{
		validateOpened ( );
			
		m_sqlCommand. setParameter (
								nIndex,
								SQLTools. createopData (
												tmValue,
												opConstants. TIME ) );
	}

	public void setTime ( int nIndex, Time  x, Calendar  cal) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setTime" );
	}

	public void setTimestamp ( int nIndex, Timestamp  x) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setTimestamp" );
	}

	public void setTimestamp ( int nIndex, Timestamp  x, Calendar  cal) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setTimestamp" );
	}

	public void setUnicodeStream ( int nIndex, InputStream  x, int length) throws SQLException
	{
		validateOpened ( );
			
		throwUnimplementedException ( "setUnicodeStream" );
	}
	private void s_e_p_a_r_a_t_o_r__________________________1 ( )
	{
		;
	}
	
	private void throwUnimplementedException ( String strMethodName ) throws SQLException
	{
		throw new SQLException (
							"com.fourd.jdbc.PreparedStatementImpl: " +
							"Method " + strMethodName + " is not implemented" );
	}
}