/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.sql.*;
import java.util.*;

import j4d.open.*;

/**
 *	Implementes ResultSetMetaData interface.
 *
 *	This class is given public access because some third-party applications ( iSQL Viewer ) for some 
 *	unknown to me reason use reflection to access result set meta data. Obviously, reflection
 *	deals with real implementation classes and not interfaces and thus requires public access.
 *	This class is made final to prevent anybody from messing with it.
 *
 */
public final class ResultSetMetaDataImpl implements ResultSetMetaData
{
	// SQLCommand object that generated this ResultSetMetaDataImpl instance.
	private SQLCommand				m_sqlCommand = null;
	
	private DatabaseMetaDataImpl		m_dbMD = null;
	
	ResultSetMetaDataImpl ( SQLCommand sqlCommand, DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		if ( sqlCommand == null )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetMetaDataImpl::ResultSetMetaDataImpl: " +
								"SQL command can not be null" );
										
		m_sqlCommand = sqlCommand;
		m_dbMD = dbMD;
	}
	
	public String getCatalogName ( int nColumn ) throws java.sql.SQLException
	{
		// throwUnsupportedException ( );
		
		return "";
	}
	
	public String getColumnClassName ( int nColumn ) throws java.sql.SQLException
	{
		SQLColumnMetaData		sqlcMD = m_sqlCommand. getSELECTColumn ( nColumn - 1 ); // does its own index check
		int						nColumnType = sqlcMD. getopType ( );
		
		if ( nColumnType == opConstants. ALPHANUMERIC )
			return String. class. getName ( );
		else if ( nColumnType == opConstants. BOOLEAN )
			return Boolean. class. getName ( );
		else if ( nColumnType == opConstants. INTEGER )
			return Short. class. getName ( );
		else if ( nColumnType == opConstants. LONGINTEGER )
			return Integer. class. getName ( );
		else if ( nColumnType == opConstants. REAL )
			return Double. class. getName ( );
		else if ( nColumnType == opConstants. TEXT )
			return String. class. getName ( );
		else if ( nColumnType == opConstants. DATE || nColumnType == opConstants. TIME )
			return java.sql.Date. class. getName ( );
		else if ( nColumnType == opConstants. BLOB )
			return byte [ ]. class. getName ( );
		else if ( nColumnType == opConstants. PICTURE )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetMetaDataImpl::getColumnClassName: " +
								"Unimplemented data type - PICTURE" );
		else if ( nColumnType == opConstants. SUBTABLE )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetMetaDataImpl::getColumnClassName: " +
								"Unimplemented data type - SUBTABLE" );
		else if ( nColumnType == opConstants. ALPHANUMERIC2 )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetMetaDataImpl::getColumnClassName: " +
								"Unimplemented data type - ALPHANUMERIC2" );
			
		throw new SQLException (
							"com.fourd.jdbc.ResultSetMetaDataImpl::getColumnClassName: " +
							"Unknown data type - " + nColumnType );
	}
	
	public int getColumnCount ( ) throws java.sql.SQLException
	{
		return m_sqlCommand. getSELECTColumnCount ( );
	}
	
	// TODO: implement properly
	public int getColumnDisplaySize ( int nColumn ) throws java.sql.SQLException
	{
		return 100;
	}
	
	public String getColumnLabel ( int nColumn ) throws java.sql.SQLException
	{
		SQLColumnMetaData		sqlcMD = m_sqlCommand. getSELECTColumn ( nColumn - 1 ); // does its own index check
		
		return	sqlcMD. getAlias ( );
	}

	public String getColumnName ( int nColumn ) throws java.sql.SQLException
	{
		SQLColumnMetaData		sqlcMD = m_sqlCommand. getSELECTColumn ( nColumn - 1 ); // does its own index check
		
		// First of all, lets figure out table's name
		String			strTableName = sqlcMD. getDBTableName ( );
		if ( strTableName == null )
			// No table name
			return sqlcMD. getDBName ( );
			
		// strTableName is not null and contains either DB name or alias
		SQLTableMetaData		sqltMD = m_sqlCommand. getTable ( strTableName );
		
		return sqltMD. getDBName ( ) + "." + sqlcMD. getDBName ( );
	}
	
	public int getColumnType ( int nColumn ) throws java.sql.SQLException
	{
		SQLColumnMetaData		sqlcMD = m_sqlCommand. getSELECTColumn ( nColumn - 1 ); // does its own index check
		String					strTableName = getTableName ( nColumn );
		SQLTableMetaData			sqltMDDB = m_dbMD. getTable ( strTableName );
		SQLColumnMetaData		sqlcMDDB = sqltMDDB. getColumnMetaData ( sqlcMD. getDBName ( ) );
		
		int						nColumnType = sqlcMDDB. getopType ( );
		
		if ( nColumnType == opConstants. ALPHANUMERIC )
			return java.sql.Types. CHAR;
		else if ( nColumnType == opConstants. BOOLEAN )
			return java.sql.Types. BIT;
		else if ( nColumnType == opConstants. INTEGER )
			return java.sql.Types. SMALLINT;
		else if ( nColumnType == opConstants. LONGINTEGER )
			return java.sql.Types. INTEGER;
		else if ( nColumnType == opConstants. REAL )
			return java.sql.Types. REAL;
		else if ( nColumnType == opConstants. TEXT )
			return java.sql.Types. VARCHAR;
		else if ( nColumnType == opConstants. DATE || nColumnType == opConstants. TIME )
			return java.sql.Types. DATE;
		else if ( nColumnType == opConstants. BLOB )
			return java.sql.Types. BLOB;
		else if ( nColumnType == opConstants. PICTURE )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetMetaDataImpl::getColumnType: " +
								"Unimplemented data type - PICTURE" );
		else if ( nColumnType == opConstants. SUBTABLE )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetMetaDataImpl::getColumnType: " +
								"Unimplemented data type - SUBTABLE" );
		else if ( nColumnType == opConstants. ALPHANUMERIC2 )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetMetaDataImpl::getColumnType: " +
								"Unimplemented data type - ALPHANUMERIC2" );
			
		throw new SQLException (
							"com.fourd.jdbc.ResultSetMetaDataImpl::getColumnType: " +
							"Unknown data type - " + nColumnType );
	}
	
	public String getColumnTypeName ( int nColumn ) throws java.sql.SQLException
	{
		SQLColumnMetaData		sqlcMD = m_sqlCommand. getSELECTColumn ( nColumn - 1 ); // does its own index check
		String					strTableName = getTableName ( nColumn );
		SQLTableMetaData			sqltMDDB = m_dbMD. getTable ( strTableName );
		SQLColumnMetaData		sqlcMDDB = sqltMDDB. getColumnMetaData ( sqlcMD. getDBName ( ) );
		
		int						nColumnType = sqlcMDDB. getopType ( );
		
		if ( nColumnType == opConstants. ALPHANUMERIC )
			return "ALPHANUMERIC";
		else if ( nColumnType == opConstants. BOOLEAN )
			return "BOOLEAN";
		else if ( nColumnType == opConstants. INTEGER )
			return "INTEGER";
		else if ( nColumnType == opConstants. LONGINTEGER )
			return "LONGINTEGER";
		else if ( nColumnType == opConstants. REAL )
			return "REAL";
		else if ( nColumnType == opConstants. TEXT )
			return "TEXT";
		else if ( nColumnType == opConstants. DATE )
			return "DATE";
		else if ( nColumnType == opConstants. TIME )
			return "TIME";
		else if ( nColumnType == opConstants. BLOB )
			return "BLOB";
		else if ( nColumnType == opConstants. PICTURE )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetMetaDataImpl::getColumnTypeName: " +
								"Unimplemented data type - PICTURE" );
		else if ( nColumnType == opConstants. SUBTABLE )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetMetaDataImpl::getColumnTypeName: " +
								"Unimplemented data type - SUBTABLE" );
		else if ( nColumnType == opConstants. ALPHANUMERIC2 )
			throw new SQLException (
								"com.fourd.jdbc.ResultSetMetaDataImpl::getColumnTypeName: " +
								"Unimplemented data type - ALPHANUMERIC2" );
			
		throw new SQLException (
							"com.fourd.jdbc.ResultSetMetaDataImpl::getColumnTypeName: " +
							"Unknown data type - " + nColumnType );
	}
	
	// TODO: implement properly
	public int getPrecision ( int nColumn ) throws java.sql.SQLException
	{
		return 10;
	}
	
	// TODO: implement properly
	public int getScale ( int nColumn ) throws java.sql.SQLException
	{
		return 10;
	}
	
	public String getSchemaName ( int nColumn ) throws java.sql.SQLException
	{
		// throwUnsupportedException ( );
		
		return "";
	}
	
	public String getTableName ( int nColumn ) throws java.sql.SQLException
	{
		SQLColumnMetaData		sqlcMD = m_sqlCommand. getSELECTColumn ( nColumn - 1 ); // does its own index check
		
		String					strTableName = sqlcMD. getDBTableName ( );
		SQLTableMetaData		sqltMD = null;
		if ( strTableName != null )
		{
			// strTableName is not null and contains either DB name or alias
			sqltMD = m_sqlCommand. getTable ( strTableName );
			
			return sqltMD. getDBName ( );
		}
		
		// Column does not have a table name and thus it is from FROM table
		return m_sqlCommand. getFROMTables ( ) [ 0 ]. getDBName ( );
	}
	
	public boolean isAutoIncrement ( int nColumn ) throws java.sql.SQLException
	{
		// throwUnsupportedException ( "isAutoIncrement" );
		
		// TODO: implement properly
		return false;
	}
	
	public boolean isCaseSensitive ( int nColumn ) throws java.sql.SQLException
	{
		/*
			TODO: implement. For all but text and alphanumeric columns should return 
			false. For text and alphanumeric columns should return false if a column
			is indexed and marked as unique, should return true.
			How do I find if a column is indexed or unique? Probably use opField.mAttributes?
		*/
		// throwUnimplementedException ( "isCaseSensitive" );
		
		return false;
	}
	
	public boolean isCurrency ( int nColumn ) throws java.sql.SQLException
	{
		//throwUnsupportedException ( "isCurrency" );

		// TODO: implement properly
		return false;
	}
	
	public boolean isDefinitelyWritable ( int nColumn ) throws java.sql.SQLException
	{
		throwUnimplementedException ( "isDefinitelyWritable" );
		
		return true;
	}
	
	public int isNullable ( int nColumn ) throws java.sql.SQLException
	{
		return columnNoNulls;
	}
	
	public boolean isReadOnly ( int nColumn ) throws java.sql.SQLException
	{
		throwUnimplementedException ( "isReadOnly" );

		return false;
	}
	
	public boolean isSearchable ( int nColumn ) throws java.sql.SQLException
	{
		throwUnimplementedException ( "isSearchable" );

		return true;
	}
	
	public boolean isSigned ( int nColumn ) throws java.sql.SQLException
	{
		throwUnimplementedException ( "isSigned" );
		
		return false;
	}
	
	public boolean isWritable ( int nColumn ) throws java.sql.SQLException
	{
		throwUnimplementedException ( "isWritable" );

		return true;
	}
	
	private void throwUnimplementedException ( String strMethodName ) throws SQLException
	{
		throw new SQLException (
							"com.fourd.jdbc.ResultSetMetaDataImpl: Method " +
							strMethodName + " is not implemented" );
	}

	private void throwUnsupportedException ( String strMethodName ) throws SQLException
	{
		throw new SQLException (
							"com.fourd.jdbc.ResultSetMetaDataImpl:" +
							" Method " +
							strMethodName +
							" is not supported" );
	}
}