/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.util.*;
import java.sql.*;

import j4d.open.*;

/**
*	Represents a row that is being created or being updated. Stores new or 
*	updated values. Values are saved to the database table when
*	Resultset. updateRow ( ) method is called.
*
*	New/updated values are stored in the hashtable. Key is a string
*	representation of an index of a column within a result set. When update
*	is performed, column index within a result set should be converted to 
*	column index within database table that is being updated.
*/
class RowCache
{
	private Hashtable					m_htCache = null;
	
	private SQLCommand				m_sqlCommand = null;
	private DatabaseMetaDataImpl		m_dbMD = null;	
	
	RowCache (
			SQLCommand sqlCommand,
			DatabaseMetaDataImpl dbMD ) throws SQLException
	{
		m_htCache = new Hashtable ( );
		
		if ( sqlCommand == null )
			throw new SQLException (
								"com.fourd.jdbc.RowCache::RowCache: " +
								"SQLCommand can not be null" );
		m_sqlCommand = sqlCommand;
		
		if ( dbMD == null )
			throw new SQLException (
								"com.fourd.jdbc.RowCache::RowCache: " +
								"DatabaseMetaDataImpl can not be null" );
		m_dbMD = dbMD;
	}
	
	/**
	*	nColumnIndex is 1-based column's index within a result set, not within a table.
	*/
	void put ( int nColumnIndex, Object objValue )
	{
		/*
			Note, that here I store column index as 0-based because I do it
			this wasy everywhere else.
		*/
		m_htCache. put ( String. valueOf ( nColumnIndex - 1 ), objValue );
	}
	
	/**
	*	Returns opFieldArray for columns of table strTableName whose changed values are 
	*	stored in this instance of RowCache.
	*	
	*	Caller of this method should always check if the resulting array has 0 length.
	*/
	opFieldArray getopFields ( String strTableName ) throws SQLException
	{
		if ( strTableName == null )
			throw new SQLException (
								"com.fourd.jdbc.RowCache::getopFields: " +
								"Table name can not be null" );
								
		int						nCachedColumnCount = getCachedColumnCount ( strTableName );
		opFieldArray			opfaResult = new opFieldArray ( nCachedColumnCount );
		if ( nCachedColumnCount == 0 )
			return opfaResult;
			
		// Get DB version of SQLTableMetaData for strTableName
		SQLTableMetaData		sqltDBMD = m_dbMD. getTable ( strTableName );
		opfaResult. mTargetTable = sqltDBMD. getNumber ( );
			
		int [ ]					sznQueryIndexes = m_sqlCommand. getSELECTColumnIndexes ( strTableName );
		SQLColumnMetaData		sqlcMD = null;
		int						nCachedIndex = 0;
		for ( int i = 0; i < sznQueryIndexes. length; i++ )
		{
			if ( m_htCache. get ( String. valueOf ( sznQueryIndexes [ i ] ) ) == null )
				// column is not cached (updated)
				continue;
				
			sqlcMD = m_sqlCommand. getSELECTColumn ( sznQueryIndexes [ i ] ); // column meta data from query
			sqlcMD = sqltDBMD. getColumnMetaData ( sqlcMD. getDBName ( ) ); // column meta data from DB
			opfaResult. mFieldArray [ nCachedIndex ]. mFieldNumber = sqlcMD. getNumber ( );
			opfaResult. mFieldArray [ nCachedIndex ]. mFieldType = sqlcMD. getopType ( );
			
			nCachedIndex++;
		}
		
		return opfaResult;
	}
	
	opDataArray getopData ( String strTableName ) throws SQLException
	{
		if ( strTableName == null )
			throw new SQLException (
								"com.fourd.jdbc.RowCache::getopData: " +
								"Table name can not be null" );
								
		int						nCachedColumnCount = getCachedColumnCount ( strTableName );
		opDataArray				opdaResult = new opDataArray ( nCachedColumnCount );
		if ( nCachedColumnCount == 0 )
			return opdaResult;
			
		// Get DB version of SQLTableMetaData for strTableName
		SQLTableMetaData		sqltDBMD = m_dbMD. getTable ( strTableName );
			
		int [ ]					sznQueryIndexes = m_sqlCommand. getSELECTColumnIndexes ( strTableName );
		SQLColumnMetaData		sqlcMD = null;
		int						nCachedIndex = 0;
		Object					objValue = null;
		for ( int i = 0; i < sznQueryIndexes. length; i++ )
		{
			if ( ( objValue = m_htCache. get ( String. valueOf ( sznQueryIndexes [ i ] ) ) ) == null )
				// column is not cached (updated)
				continue;
				
			sqlcMD = m_sqlCommand. getSELECTColumn ( sznQueryIndexes [ i ] ); // column meta data from query
			sqlcMD = sqltDBMD. getColumnMetaData ( sqlcMD. getDBName ( ) ); // column meta data from DB
			opdaResult. mDataArray [ nCachedIndex ] = SQLTools.
															createopData (
																	objValue,
																	sqlcMD. getopType ( ) );
			
			nCachedIndex++;
		}
		
		return opdaResult;
	}
	
	/**
	*	Returns amount of columns from the given table strTableName that are being cached in this
	*	RowCache instance.
	*/
	private int getCachedColumnCount ( String strTableName ) throws SQLException
	{
		int [ ]		sznQueryIndexes = m_sqlCommand. getSELECTColumnIndexes ( strTableName );
		int			nResult = 0;
		for ( int i = 0; i < sznQueryIndexes. length; i++ )
			if ( m_htCache. get ( String. valueOf ( sznQueryIndexes [ i ] ) ) != null )
				nResult++;
				
		return nResult;
	}
	
	/**
	*	Input parameter nColumn is 0-based.
	*/
	boolean columnIsCached ( int nColumn )
	{
		return m_htCache. get ( String. valueOf ( nColumn ) ) != null;
	}
	
	/**
	*	Input parameter nColumn is 0-based.
	*/
	Object get ( int nColumn )
	{
		return m_htCache. get ( String. valueOf ( nColumn ) );
	}
	
	int size ( )
	{
		return m_htCache. size ( );
	}
}