/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

/**
*	Parses SQL tokens into a SQL command.
*/
class SQLParser
{
	// Hide default constructor
	private SQLParser ( )
	{
	
	}
	
	/*SQLParser ( SQLTokenizer sqlTokenizer ) throws IllegalArgumentException
	{
		if ( sqlTokenizer == null )
			throw new IllegalArgumentException (
						"com.fourd.jdbc.SQLParser.SQLParser: SQLTokenizer must not be null" );
						
		m_sqlTokenizer = sqlTokenizer;
	}
	*/
	
	/**
	*	Parses sql query and returns SQLCommand object.
	*	Throws an exception if sql query is invalid or not supported.
	*/
	static SQLCommand parse ( String strSQL ) throws IllegalArgumentException, Exception
	{
		if ( strSQL == null )
			throw new IllegalArgumentException (
						"com.fourd.jdbc.SQLParser.SQLParser.parse: SQL string must not be null" );
						
		SQLTokenizer	sqlTokenizer = new SQLTokenizer ( new StringBuffer ( strSQL ) );

		if ( !sqlTokenizer. hasMoreTokens ( ) )
			throw new Exception ( "com.fourd.jdbc.SQLParser.parse: SQL query does not contain any commands" );
			
		int [ ]			sznParameterIndex = new int [ ] { 0 };

		SQLCommand		sqlCommand = null;
		String			strToken = sqlTokenizer. getNextToken ( );
		
		// Lets find out the type of command
		if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_SELECT ) )
		{
			sqlCommand = new SQLCommand ( SQLCommand. CMD_TYPE_SELECT );
			
			// Lets see if there is ALL or DISTINCT keyword
			if ( !sqlTokenizer. hasMoreTokens ( ) )
				throw new Exception ( "com.fourd.jdbc.SQLParser.parse: SELECT query is incomplete" );
				
			strToken = sqlTokenizer. getNextToken ( );
			if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_DISTINCT ) )
				// TODO: find out if there is a way in 4D to select distinct records
				throw new Exception ( "com.fourd.jdbc.SQLParser.parse: DISTINCT keyword is not supported" );
			else if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_ALL ) )
				// Currently ALL is the default behavior. Do nothing.
				;
			else
				// Oops, started to read field list. Lets roll back one token.
				sqlTokenizer. rollBack ( 1 );
				
			// Lets parse column list
			parseSELECTColumnList ( sqlTokenizer, sqlCommand );
			
			if ( !sqlTokenizer. hasMoreTokens ( ) )
				throw new Exception ( "com.fourd.jdbc.SQLParser.parse: FROM keyword is missing" );
				
			strToken = sqlTokenizer. getNextToken ( );
			if ( !strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_FROM ) )
				throw new Exception (
								"com.fourd.jdbc.SQLParser.parse: Unexpected token - " +
								strToken +
								". FROM keyword was expected instead." );
								
			parseFROMTableList ( sqlTokenizer, sqlCommand );
			
			if ( !sqlTokenizer. hasMoreTokens ( ) )
				// We are done. This was SELECT .. FROM ... command.
				return sqlCommand;
			
			// JOIN clause. Currently we only support LEFT JOIN
			strToken = sqlTokenizer. getNextToken ( );
			if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_GO ) )
				return sqlCommand;
				
			if (	strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_LEFT ) )
			{
				parseJOIN ( sqlTokenizer, sqlCommand );
				
				if ( !sqlTokenizer. hasMoreTokens ( ) )
					// We are done. This was SELECT .. FROM ... JOIN command.
					return sqlCommand;

				strToken = sqlTokenizer. getNextToken ( );
				if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_GO ) )
					return sqlCommand;
			}
			
			// WHERE clause
			if (	strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_WHERE ) ||
				strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_WHERE4D ) )
			{
				parseWHERE ( sqlTokenizer, sqlCommand, sznParameterIndex );
				sqlCommand. setWHEREType ( strToken. toUpperCase ( ) );
				
				if ( !sqlTokenizer. hasMoreTokens ( ) )
					// We are done. This was SELECT .. FROM ... WHERE(4D) command.
					return sqlCommand;

				strToken = sqlTokenizer. getNextToken ( );
				if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_GO ) )
					return sqlCommand;
			}
				
			// Now we expect ORDER BY clause
			if ( !strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_ORDER ) )
				throw new Exception (
								"com.fourd.jdbc.SQLParser.parse: Unexpected token - " +
								strToken +
								". ORDER keyword was expected instead." );
			
			if ( !sqlTokenizer. hasMoreTokens ( ) )
				throw new Exception (
								"com.fourd.jdbc.SQLParser.parse: Unexpected end of query - " +
								"BY keyword was expected instead." );
				
			strToken = sqlTokenizer. getNextToken ( );
			if ( !strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_BY ) )
				throw new Exception (
								"com.fourd.jdbc.SQLParser.parse: Unexpected token - " +
								strToken +
								". BY keyword was expected instead." );
								
			parseORDERBYColumnList ( sqlTokenizer, sqlCommand );

			// Currently all we support is SELECT and FROM and ORDER By
			if (	sqlTokenizer. hasMoreTokens ( ) &&
					!( strToken = sqlTokenizer. getNextToken ( ) ). toUpperCase ( ). equals (
																		SQLCommand. SQL_KEYWORD_GO ) )
				throw new Exception (
								"com.fourd.jdbc.SQLParser.parse: Unexpected token - " +
								sqlTokenizer. getNextToken ( ) +
								". Expected end of query." );
		} // end of  if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_SELECT ) )
		else if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_DELETE ) )
		{
			sqlCommand = new SQLCommand ( SQLCommand. CMD_TYPE_DELETE );
			parseDELETE ( sqlTokenizer, sqlCommand );
			
			if ( !sqlTokenizer. hasMoreTokens ( ) )
				// We are done. This was simple DELETE .. FROM ... command.
				return sqlCommand;

			strToken = sqlTokenizer. getNextToken ( ); // Must be WHERE[4D] keyword
			
			// WHERE[4D] clause
			if (	strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_WHERE ) ||
				strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_WHERE4D ) )
			{
				parseWHERE ( sqlTokenizer, sqlCommand, sznParameterIndex );
				sqlCommand. setWHEREType ( strToken. toUpperCase ( ) );
				
				if ( !sqlTokenizer. hasMoreTokens ( ) )
					// We are done. This was DELETE .. FROM ... WHERE(4D) command.
					return sqlCommand;
				else
					strToken = sqlTokenizer. getNextToken ( );
			}

			throw new Exception ( "com.fourd.jdbc.SQLParser.parse: Unexpected token - " + strToken );
		} // end of else if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_DELETE ) )
		else if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_INSERT ) )
		{
			sqlCommand = new SQLCommand ( SQLCommand. CMD_TYPE_INSERT );
			parseINSERT ( sqlTokenizer, sqlCommand, sznParameterIndex );
			
			if ( sqlTokenizer. hasMoreTokens ( ) )
			{
				strToken = sqlTokenizer. getNextToken ( );
				
				throw new Exception (
									"com.fourd.jdbc.SQLParser.parse: Unexpected token - " +
									strToken + ". Expected end of query instead." );
			}
		} // end of else if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_INSERT ) )
		else if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_UPDATE ) )
		{
			sqlCommand = new SQLCommand ( SQLCommand. CMD_TYPE_UPDATE );
			parseUPDATE ( sqlTokenizer, sqlCommand );
			parseUPDATEList ( sqlTokenizer, sqlCommand, sznParameterIndex );
			if ( !sqlTokenizer. hasMoreTokens ( ) )
				// This is a short UPDATE ... SET ... command that updates the whole table.
				return sqlCommand;
				
			strToken = sqlTokenizer. getNextToken ( ); // Must be WHERE[4D] keyword
			
			// WHERE[4D] clause
			if (	strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_WHERE ) ||
					strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_WHERE4D ) )
			{
				parseWHERE ( sqlTokenizer, sqlCommand, sznParameterIndex );
				sqlCommand. setWHEREType ( strToken. toUpperCase ( ) );
				
				if ( !sqlTokenizer. hasMoreTokens ( ) )
					// We are done. This was UPDATE .. SET ... WHERE[4D] command.
					return sqlCommand;
				else
					strToken = sqlTokenizer. getNextToken ( );
			}

			throw new Exception ( "com.fourd.jdbc.SQLParser.parse: Unexpected token - " + strToken );
		}
		else if ( strToken. toUpperCase ( ). equals ( "{" ) )
		{
			sqlCommand = new SQLCommand ( SQLCommand. CMD_TYPE_STORED_PROCEDURE );
			parseCALL ( sqlTokenizer, sqlCommand );
		}
		else
			throw new Exception (
							"com.fourd.jdbc.SQLParser.parse: Unsupported SQL command - " +
							strToken );
							
		return sqlCommand;
	}
	
	/**
	*	Parses a list of table columns separated by commas.
	*	sqlTokenizer is used to extract table columns. sqlCommand will be updated with the columns
	*	information when the method finishes. List of columns is expected to be formatted as in
	*	SELECT ALL comma_separated_list_of_columns_or_* ... . An exception is thrown if the query is
	*	invalid.
	*
	*	Uses SQLTokenizer. hasMoreTokens ( ) method a lot. I could just use getNextToken and whenever a token
	*	is expected but is in fact missing SQLTokenizer would throw an exception. This, however, would
	*	produce a generic "No more tokens" message whereas when I use hasMoreTokens I can generate more
	*	specific message about what exactly is wrong with the query.
	*
	*	TODO: This method's implementation contains blocks of code that are variations of the same thing. Should be easy 
	*	to create a few helper functions to simplify it.
	*/
	static private void parseSELECTColumnList ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	{
		if ( !sqlTokenizer. hasMoreTokens ( ) )
			throw new Exception ( "com.fourd.jdbc.SQLParser.parseSELECTColumnList: Query is incomplete, there are no columns to be selected" );
			
		String		strToken = sqlTokenizer. getNextToken ( );
		if ( strToken. equals ( "*" ) )
		{
			// Select all fields
			sqlCommand. addSELECTColumn ( new SQLColumnMetaData ( strToken ) );
			
			return;
		}
		
		sqlTokenizer. rollBack ( 1 );
		
		SQLColumnMetaData		sqlcMD = null;
		String					strColumnName = null;
		String					strTableName = null;
		/**
		*	One pass of the following while loop deals with exactly one column:
		*		[TableName].ColumnName [ AS ColumnAlias ] [,]
		*	In the example above [ and ] are used to indicate optional parts.
		*/
		while ( sqlTokenizer. hasMoreTokens ( ) )
		{
			strColumnName = sqlTokenizer. getNextToken ( );
			
			if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_FROM ) )
			{
				// FROM section starts and thus all columns are parsed. Lets roll back and return.
				sqlTokenizer. rollBack ( 1 );
				
				return;
			}
			
			if ( !sqlTokenizer. hasMoreTokens ( ) ) // This is just a column name without a table name or alias
			{
				validateColumnName ( strColumnName );
				
				sqlCommand. addSELECTColumn (
									new SQLColumnMetaData (
											SQLTools. removeEnclosers ( strColumnName, '[', ']' ) ) );
												
				return;
			}
				
			// There are still more tokens.
			strToken = sqlTokenizer. getNextToken ( );
			if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_FROM ) )
			{
				validateColumnName ( strColumnName );

				sqlcMD = new SQLColumnMetaData ( );
				sqlcMD. setDBName (
							SQLTools. removeEnclosers ( strColumnName, '[', ']' ) );
				sqlCommand. addSELECTColumn ( sqlcMD );
				
				// FROM section starts and thus all columns are parsed. Lets roll back and return.
				sqlTokenizer. rollBack ( 1 );
				
				return;
			}
			else if ( strToken. equals ( "." ) ) // A separator between table name and column name.
			{
				// strColumnName in fact contains table name
				strTableName = strColumnName;
				
				validateTableName ( strTableName );
				
				if ( !sqlTokenizer. hasMoreTokens ( ) ) // A column name is expected but is missing.
					throw new Exception (
								"com.fourd.jdbc.SQLParser.parseSELECTColumnList: Column name is missing" );
								
				strColumnName = sqlTokenizer. getNextToken ( );
				
				validateColumnName ( strColumnName );
				
				sqlcMD = new SQLColumnMetaData ( );
				sqlcMD. setDBName (
							SQLTools. removeEnclosers ( strColumnName, '[', ']' ) );
				sqlcMD. setTableAlias (
							SQLTools. removeEnclosers ( strTableName, '[', ']' ) );
				if ( !sqlTokenizer. hasMoreTokens ( ) ) // There is no AS keyword or alias
				{
					sqlCommand. addSELECTColumn ( sqlcMD );
												
					return;
				}
				
				strToken = sqlTokenizer. getNextToken ( );
				if ( strToken. equals ( "," ) ) // There are more columns
				{
					sqlCommand. addSELECTColumn ( sqlcMD );
					
					continue;
				}
				else if ( SQLTools. isValidColumnAlias ( strToken ) )
				{
					sqlcMD. setAlias (
							SQLTools. removeEnclosers ( strToken, '"', '"' ) );
					sqlCommand. addSELECTColumn ( sqlcMD );
				
					if ( !sqlTokenizer. hasMoreTokens ( ) )
						return;
					
					strToken = sqlTokenizer. getNextToken ( );
					if ( strToken. equals ( "," ) ) // There are more columns
						continue;
						
					// FROM section should follow but this s a subject for another method
					sqlTokenizer. rollBack ( 1 );
					
					return;
				}
				else if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_AS ) )
				{
					// We have an AS keyword so lets try to extract an alias for a column
					if ( !sqlTokenizer. hasMoreTokens ( ) )
						throw new Exception (
										"com.fourd.jdbc.SQLParser.parseSELECTColumnList: Query is not complete - " +
										" column alias is missing" );
									
					strToken = sqlTokenizer. getNextToken ( );

					// Lets make sure that alias is valid
					if ( !SQLTools. isValidColumnAlias ( strToken ) )
						throw new Exception (
										"com.fourd.jdbc.SQLParser.parseSELECTColumnList: Invalid column alias - " +
										strToken );
					
					sqlcMD. setAlias (
								SQLTools. removeEnclosers ( strToken, '"', '"' ) );
					sqlCommand. addSELECTColumn ( sqlcMD );
				
					if ( !sqlTokenizer. hasMoreTokens ( ) )
						return;
					
					strToken = sqlTokenizer. getNextToken ( );
					if ( strToken. equals ( "," ) ) // There are more columns
						continue;
						
					// FROM section should follow but this s a subject for another method
					sqlTokenizer. rollBack ( 1 );
					
					return;
				}
				else
				{
					sqlCommand. addSELECTColumn ( sqlcMD );
					
					// FROM section should follow but this s a subject for another method
					sqlTokenizer. rollBack ( 1 );
					
					return;
				}
			}
			else if ( strToken. equals ( "," ) ) // More columns should follow
			{
				validateColumnName ( strColumnName );
				
				sqlcMD = new SQLColumnMetaData ( );
				sqlcMD. setDBName (
							SQLTools. removeEnclosers ( strColumnName, '[', ']' ) );
				sqlCommand. addSELECTColumn ( sqlcMD );

				continue;
			}
			else if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_AS ) ) // An alias for a column name should follow
			{
				validateColumnName ( strColumnName );

				// We have an AS keyword so lets try to extract an alias for a column
				if ( !sqlTokenizer. hasMoreTokens ( ) )
					throw new Exception (
									"com.fourd.jdbc.SQLParser.parseSELECTColumnList: Query is not complete - " +
									" column alias is missing" );
									
				strToken = sqlTokenizer. getNextToken ( );

				// Lets make sure that alias is valid
				if ( !SQLTools. isValidColumnAlias ( strToken ) )
					throw new Exception (
									"com.fourd.jdbc.SQLParser.parseSELECTColumnList: Invalid column alias - " +
									strToken );
									
				sqlcMD = new SQLColumnMetaData ( );
				sqlcMD. setDBName (
							SQLTools. removeEnclosers ( strColumnName, '[', ']' ) );
				sqlcMD. setAlias (
							SQLTools. removeEnclosers ( strToken, '"', '"' ) );
				sqlCommand. addSELECTColumn ( sqlcMD );

				if ( !sqlTokenizer. hasMoreTokens ( ) )
					return;
					
				strToken = sqlTokenizer. getNextToken ( );
				if ( strToken. equals ( "," ) ) // There are more columns
					continue;
						
				// FROM section should follow but this s a subject for another method
				sqlTokenizer. rollBack ( 1 );
					
				return;
			}
			else if ( SQLTools. isValidColumnAlias ( strToken ) ) // Valid alias just without an AS keyword
			{
				validateColumnName ( strColumnName );

				sqlcMD = new SQLColumnMetaData ( );
				sqlcMD. setDBName (
							SQLTools. removeEnclosers ( strColumnName, '[', ']' ) );
				sqlcMD. setAlias (
							SQLTools. removeEnclosers ( strToken, '"', '"' ) );
				sqlCommand. addSELECTColumn ( sqlcMD );

				if ( !sqlTokenizer. hasMoreTokens ( ) )
					return;
					
				strToken = sqlTokenizer. getNextToken ( );
				if ( strToken. equals ( "," ) ) // There are more columns
					continue;
						
				// FROM section should follow but this s a subject for another method
				sqlTokenizer. rollBack ( 1 );
					
				return;
			}
			else
				throw new Exception (
							"com.fourd.jdbc.SQLParser.parseSELECTColumnList: Unexpected token - " +
							strToken );
		} // end of while ( sqlTokenizer. hasMoreTokens ( ) )
	} // end of static private void parseSELECTColumnList ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	
	private static void parseFROMTableList ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	{
		if ( !sqlTokenizer. hasMoreTokens ( ) )
			throw new Exception ( "com.fourd.jdbc.SQLParser.parseFROMTableList: Query is incomplete, there are no tables to select from" );
			
		String		strToken = null;
		String		strTableName = null;
		/**
			A single pass of the following loop deals with exactly one table declaration in the form provided below:
				TableName [[AS] TableAlias] [,]
		*/
		while ( sqlTokenizer. hasMoreTokens ( ) )
		{
			strTableName = sqlTokenizer. getNextToken ( );
			validateTableName ( strTableName );
			
			if ( !sqlTokenizer. hasMoreTokens ( ) ) // This is really a corner case and should never happen unless the originar query is not valid
			{
				sqlCommand. addFROMTable (
								new SQLTableMetaData (
										SQLTools. removeEnclosers ( strTableName, '[', ']' ) ) );
				
				return;
			}
			
			// There are more tokens
			strToken = sqlTokenizer. getNextToken ( );
			
			// Next valid token is either a comma, AS keyword, table alias or SQL keyword
			if ( strToken. equals ( "," ) ) // Just a table name
			{
				sqlCommand. addFROMTable (
								new SQLTableMetaData (
										SQLTools. removeEnclosers ( strTableName, '[', ']' ) ) );
				
				continue;
			}
			else if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_AS ) ) // An alias for a column name should follow
			{
				if ( !sqlTokenizer. hasMoreTokens ( ) )
					throw new Exception (
									"com.fourd.jdbc.SQLParser.parseFROMTableList: " +
									"Query is incomplete - table alias is missing" );
									
				strToken = sqlTokenizer. getNextToken ( );
				if ( !SQLTools. isValidTableAlias ( strToken ) )
					throw new Exception (
									"com.fourd.jdbc.SQLParser.parseFROMTableList: " +
									"Invalid table alias - " + strToken );
									
				SQLTableMetaData		sqltMD = new SQLTableMetaData ( );
				sqltMD. setDBName ( SQLTools. removeEnclosers ( strTableName, '[', ']' ) );
				sqltMD. setAlias ( strToken );
				sqlCommand. addFROMTable ( sqltMD );

				if ( !sqlTokenizer. hasMoreTokens ( ) )
					return;
					
				strToken = sqlTokenizer. getNextToken ( );
				if ( strToken. equals ( "," ) ) // There are still more tables to parse
					continue;
				
				// This is the end of FROM table list
				sqlTokenizer. rollBack ( 1 );
				
				return;
			}
			else if ( SQLTools. isValidTableAlias ( strToken ) )
			{
				SQLTableMetaData		sqltMD = new SQLTableMetaData ( );
				sqltMD. setDBName ( SQLTools. removeEnclosers ( strTableName, '[', ']' ) );
				sqltMD. setAlias ( strToken );
				sqlCommand. addFROMTable ( sqltMD );
				
				if ( !sqlTokenizer. hasMoreTokens ( ) )
					return;
					
				strToken = sqlTokenizer. getNextToken ( );
				if ( strToken. equals ( "," ) ) // There are still more tables to parse
					continue;
				
				// This is the end of FROM table list
				sqlTokenizer. rollBack ( 1 );
				
				return;
			}
			else
			{
				sqlCommand. addFROMTable (
								new SQLTableMetaData (
										SQLTools. removeEnclosers ( strTableName, '[', ']' ) ) );
				
				sqlTokenizer. rollBack ( 1 );

				return;
			}
		} // end of while ( sqlTokenizer. hasMoreTokens ( ) )
	} // end of private static void parseFROMTableList ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	
	/**
	*	Parses ORDER BY column list and updates sqlCommand object.
	*	Items in the ORDER BY list are separated by comma. Each item is of the following form:
	*		[TableName.]ColumnName [DESC|ASC]
	*/
	private static void parseORDERBYColumnList ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	{
		if ( !sqlTokenizer. hasMoreTokens ( ) )
			throw new Exception ( "com.fourd.jdbc.SQLParser.parseORDERBYColumnList: Query is incomplete, there are no columns to order by" );
			
		String					strToken = null;
		String					strTableName = null;
		String					strColumnName = null;
		SQLColumnMetaData		sqlcMD = null;
		while ( sqlTokenizer. hasMoreTokens ( ) )
		{
			strToken = sqlTokenizer. getNextToken ( );
			
			if ( !sqlTokenizer. hasMoreTokens ( ) ) // strToken contains column name that should be sorted as ASC
			{
				validateColumnName ( strToken );
				
				sqlcMD = new SQLColumnMetaData (
										SQLTools. removeEnclosers ( strToken, '[', ']' ) );
				sqlcMD. setOrderBy ( SQLColumnMetaData. ORDER_BY_ASC );
				sqlCommand. addORDERBYColumn ( sqlcMD );
				
				return;
			}
			
			/*
				We have some more tokens to parse.
				Next token could be one of the following: a period that separates table name from column name,
				ASC or DESC sql keywords, a comma, another sql keyword.
			*/
			strColumnName = strToken;
			strToken = sqlTokenizer. getNextToken ( );
			if ( strToken. equals ( "." ) )
			{
				// TableName.ColumnName case. Thus, strColumnName contains table name. Lets fix this.
				strTableName = strColumnName;
				validateTableName ( strTableName );
				
				if ( !sqlTokenizer. hasMoreTokens ( ) )
					throw new Exception (
									"com.fourd.jdbc.SQLParser.parseORDERBYColumnList: " +
									"Query is incomplete, ORDER BY column name is missing" );
					
				strColumnName = sqlTokenizer. getNextToken ( );
				validateColumnName ( strColumnName );
				
				sqlcMD = new SQLColumnMetaData (
										SQLTools. removeEnclosers ( strColumnName, '[', ']' ) );
				sqlcMD. setTableAlias ( SQLTools. removeEnclosers ( strTableName, '[', ']' ) );
				
				if ( !sqlTokenizer. hasMoreTokens ( ) )
				{
					sqlcMD. setOrderBy ( SQLColumnMetaData. ORDER_BY_ASC );
					sqlCommand. addORDERBYColumn ( sqlcMD );
				
					return;
				}
				
				strToken = sqlTokenizer. getNextToken ( );
				// Now strToken is either comma, ASC|DESC or another sql keyword
				
				if ( strToken. equals ( "," ) )
					if ( !sqlTokenizer. hasMoreTokens ( ) )
						throw new Exception (
										"com.fourd.jdbc.SQLParser.parseORDERBYColumnList: " +
										"Query is incomplete, ORDER BY column name is expected" );
					else // there are still more columns to parse
					{
						sqlCommand. addORDERBYColumn ( sqlcMD );
						continue;
					}
						
				if (
					!strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_ASC ) &&
					!strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_DESC ) )
				{
					// The next part of the SELECT query follows
					sqlTokenizer. rollBack ( 1 );
					
					sqlCommand. addORDERBYColumn ( sqlcMD );
					
					return;
				}
				
				if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_ASC ) )
					sqlcMD. setOrderBy ( SQLColumnMetaData. ORDER_BY_ASC );
				else
					sqlcMD. setOrderBy ( SQLColumnMetaData. ORDER_BY_DESC );
					
				if ( !sqlTokenizer. hasMoreTokens ( ) )
				{
					sqlCommand. addORDERBYColumn ( sqlcMD );
					
					return;
				}
					
				strToken = sqlTokenizer. getNextToken ( );
				if ( strToken. equals ( "," ) )
					if ( !sqlTokenizer. hasMoreTokens ( ) )
						throw new Exception (
										"com.fourd.jdbc.SQLParser.parseORDERBYColumnList: " +
										"Query is incomplete, ORDER BY column name is expected" );
					else // there are still more columns to parse
					{
						sqlCommand. addORDERBYColumn ( sqlcMD );
						continue;
					}
						
				// The next part of the SELECT query follows
				sqlTokenizer. rollBack ( 1 );
				sqlCommand. addORDERBYColumn ( sqlcMD );
				
				return;
			} // end of if ( strToken. equals ( "." ) )
			else if (
				strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_ASC ) ||
				strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_DESC ) )
			{
				validateColumnName ( strColumnName );
				sqlcMD = new SQLColumnMetaData (
										SQLTools. removeEnclosers ( strColumnName, '[', ']' ) );
			
				if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_ASC ) )
					sqlcMD. setOrderBy ( SQLColumnMetaData. ORDER_BY_ASC );
				else
					sqlcMD. setOrderBy ( SQLColumnMetaData. ORDER_BY_DESC );

				sqlCommand. addORDERBYColumn ( sqlcMD );

				if ( !sqlTokenizer. hasMoreTokens ( ) )
					return;
					
				strToken = sqlTokenizer. getNextToken ( );
				if ( strToken. equals ( "," ) )
					if ( !sqlTokenizer. hasMoreTokens ( ) )
						throw new Exception (
										"com.fourd.jdbc.SQLParser.parseORDERBYColumnList: " +
										"Query is incomplete, ORDER BY column name is expected" );
					else // there are still more columns to parse
						continue;

				// The next part of the SELECT query follows
				sqlTokenizer. rollBack ( 1 );
				
				return;
			}
			else if ( strToken. equals ( "," ) )
			{
				validateColumnName ( strColumnName );
				
				sqlcMD = new SQLColumnMetaData (
										SQLTools. removeEnclosers ( strColumnName, '[', ']' ) );
				sqlcMD. setOrderBy ( SQLColumnMetaData. ORDER_BY_ASC );
				
				sqlCommand. addORDERBYColumn ( sqlcMD );

				if ( !sqlTokenizer. hasMoreTokens ( ) )
					throw new Exception (
									"com.fourd.jdbc.SQLParser.parseORDERBYColumnList: " +
									"Query is incomplete, ORDER BY column name is expected" );
									
				continue;
			}
			else // This is the end of the ORDER BY list.
			{
				validateColumnName ( strColumnName );
				
				sqlcMD = new SQLColumnMetaData (
										SQLTools. removeEnclosers ( strColumnName, '[', ']' ) );
				sqlcMD. setOrderBy ( SQLColumnMetaData. ORDER_BY_ASC );
				
				sqlCommand. addORDERBYColumn ( sqlcMD );

				// The next part of the SELECT query follows
				sqlTokenizer. rollBack ( 1 );
									
				return;
			}
		} // end of while ( sqlTokenizer. hasMoreTokens ( ) )
	} // end of private static void parseORDERBYColumnList ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	
	/*
	*	Parses WHERE(4D) clause.
	*	WHERE clause can contain only boolean expressions connected by either AND or OR. There is no parenthesis support.
	*	Currently supported boolean expressions:
	*		[TableName.]ColumnName =|<>|<|>|<=|>= Value, Value := Number|String|Date
	*/
	private static void parseWHERE (
								SQLTokenizer sqlTokenizer,
								SQLCommand sqlCommand,
								int [ ] sznParameterIndex ) throws Exception
	{
		if ( !sqlTokenizer. hasMoreTokens ( ) )
			throw new Exception ( "com.fourd.jdbc.SQLParser::parseWHERE: Query is incomplete, there are no boolean expressions" );
			
		String					strToken = null;
		String					strTableName = null;
		String					strColumnName = null;
		String					strComparisonOperator = null;
		Object					objValue = null;
		SQLColumnMetaData		sqlcMD = null;
		BooleanExpression			bExpression = null;
		while ( sqlTokenizer. hasMoreTokens ( ) )
		{
			strToken = sqlTokenizer. getNextToken ( ); // Either a table or column name
			
			// There has to be more tokens
			sqlTokenizer. validateNextToken ( );
			if ( sqlTokenizer. peekAtNextToken ( ). equals ( "." ) )
			{
				// strToken contains table name
				validateTableName ( strToken );
				strTableName = strToken;
				
				sqlTokenizer. getNextToken ( ); // skip period "." that separates table alias from column name
				sqlTokenizer. validateNextToken ( ); // A column name should follow.
				strColumnName = sqlTokenizer. getNextToken ( );
				validateColumnName ( strColumnName );
			}
			else
			{
				// strToken contains column name
				strColumnName = strToken;
				validateColumnName ( strColumnName );
			}
			
			sqlcMD = new SQLColumnMetaData (
										SQLTools. removeEnclosers ( strColumnName, '[', ']' ) );
			if ( strTableName != null )
				sqlcMD. setTableAlias ( SQLTools. removeEnclosers ( strTableName, '[', ']' ) );
				
			// There has to be a coparison operator
			sqlTokenizer. validateNextToken ( );
			strToken = sqlTokenizer. getNextToken ( );
			BooleanExpression. validateComparisonOperator ( strToken );
			strComparisonOperator = strToken;
			
			/*
				Now it is time to read the value to compare to.
				A value can be
					a ?
					a number,
					a string enclosed in single quotes ('),
					a date enclosed in single quotes ('),
					boolean keywords TRUE or FALSE
			*/
			// TODO: when the parsing is complete, values should be validated to be of the same type as their columns and are actually valid themselves.
			sqlTokenizer. validateNextToken ( );
			strToken = sqlTokenizer. getNextToken ( );
			if ( strToken. equals ( "?" ) )
			{
				sqlCommand. addParameter ( null );
				objValue = new Parameter ( sznParameterIndex [ 0 ] );
				sznParameterIndex [ 0 ] ++;
			}
			else if ( SQLTools. isString ( strToken ) )
				objValue = SQLTools. removeEnclosers ( strToken, '\'', '\'' );
			else if (	strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_TRUE ) ||
					strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_FALSE ) )
				objValue = strToken. toUpperCase ( );
			else // Value is a number
			{
				// Complete number is read in the SQLTokenizer
				objValue = strToken;
			}
			
			bExpression = new BooleanExpression ( sqlcMD, strComparisonOperator, objValue );
			
			if ( !sqlTokenizer. hasMoreTokens ( ) )
			{
				sqlCommand. addWHEREExpression ( bExpression );
				
				return;
			}
				
			strToken = sqlTokenizer. getNextToken ( );
			if (	!strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_AND ) &&
				!strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_OR ) )
			{
				sqlTokenizer. rollBack ( 1 );
				sqlCommand. addWHEREExpression ( bExpression );
				
				return;
			}
			
			bExpression. setLogicalOperator ( strToken. toUpperCase ( ) );
			sqlCommand. addWHEREExpression ( bExpression );
			
			// Lets make sure that the query does not end with AND or OR
			sqlTokenizer. validateNextToken ( );
		} // end of while ( sqlTokenizer. hasMoreTokens ( ) )
	}
	
	/**
	*	Parses JOIN clause(s). Currently we only support LEFT JOIN. We support multi-table and multi-column joins.
	*	When this method is called, LEFT keyword is already consumed.
	*	
	*	JOIN_clause := [1...n] JOIN_expression
	*	JOIN_expression := LEFT JOIN JOIN_conditions
	*	JOIN_conditions := JOIN_condition [ 0...n ] AND JOIN_condition
	*	JOIN_condition := TableName [[AS] TableAlias] ON [TableAlials.]ColumnName = [TableAlials.]ColumnName
	*/
	private static void parseJOIN ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	{
		sqlTokenizer. validateNextToken ( SQLCommand. SQL_KEYWORD_JOIN );
		sqlTokenizer. getNextToken ( ); // consume JOIN
		
		sqlTokenizer. validateNextToken ( ); // Lets make sure that LEFT JOIN clause is not empty
		
		String					strToken = null;
		String					strTableName = null;
		String					strTableAlias = null;
		
		// This is the beginning of JOIN_condition and thus strToken contains table name
		strTableName = sqlTokenizer. getNextToken ( );
		validateTableName ( strTableName );
			
		SQLTableMetaData			sqltMDJoin = new SQLTableMetaData (
														SQLTools. removeEnclosers ( strTableName, '[', ']' ) );

		sqlTokenizer. validateNextToken ( );
		strToken = sqlTokenizer. peekAtNextToken ( );
		/*
			Now we can have:
				optional AS keyword,
				table alias,
				ON keyword
		*/
		if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_AS ) )
		{
			sqlTokenizer. getNextToken ( ); // consume AS
			
			// Table alias should follow
			sqlTokenizer. validateNextToken ( );
			strTableAlias = sqlTokenizer. getNextToken ( );
			validateTableAlias ( strTableAlias );
		}
		else if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_ON ) )
			; // do nothing
		else	
		{
			// We have a table alias without optional AS keyword
			strTableAlias = sqlTokenizer. getNextToken ( );
			validateTableAlias ( strTableAlias );
		}
		
		if ( strTableAlias != null )
			sqltMDJoin. setAlias ( SQLTools. removeEnclosers ( strTableAlias, '[', ']' ) );
			
		// Now we expect ON keyword
		sqlTokenizer. validateNextToken ( SQLCommand. SQL_KEYWORD_ON );
		sqlTokenizer. getNextToken ( ); // consume ON keyword
		
		String					strLeftTableAlias = null;
		String					strRightTableAlias = null;
		String					strLeftColumnName = null;
		String					strRightColumnName = null;
		SQLColumnMetaData		sqlcMDLeft = null;
		SQLColumnMetaData		sqlcMDRight = null;
		while ( sqlTokenizer. hasMoreTokens ( ) )
		{
			// Now we expect [TableAlials.]ColumnName = [TableAlials.]ColumnName [AND]
			// Lets parse left part of the expression
			
			sqlTokenizer. validateNextToken ( );
			strToken = sqlTokenizer. getNextToken ( ); // Now strToken contains either table alias or column name
			
			sqlTokenizer. validateNextToken ( );
			if ( sqlTokenizer. peekAtNextToken ( ). equals ( "." ) )
			{
				// strToken, in fact, contains table alias and column name should follow
				strLeftTableAlias = strToken;
				validateTableAlias ( strLeftTableAlias );
				sqlTokenizer. getNextToken ( ); // consume period '.'
				
				// Extract column name
				sqlTokenizer. validateNextToken ( );
				strLeftColumnName = sqlTokenizer. getNextToken ( );
				validateColumnName ( strLeftColumnName );
			}
			else
				// strToken, in fact, contains column name and '=' should follow
				strLeftColumnName = strToken;
			
			sqlcMDLeft = new SQLColumnMetaData (
										SQLTools. removeEnclosers ( strLeftColumnName, '[', ']' ) );
			if ( strLeftTableAlias != null )
				sqlcMDLeft. setTableAlias (
									SQLTools. removeEnclosers ( strLeftTableAlias, '[', ']' ) );
			
			sqlTokenizer. validateNextToken ( "=" );
			sqlTokenizer. getNextToken ( ); // consume '='
			
			// Now we expect [TableAlials.]ColumnName [AND]
			// Lets parse right part of the expression
			sqlTokenizer. validateNextToken ( );
			strToken = sqlTokenizer. getNextToken ( ); // Now strToken contains either table alias or column name
			
			sqlTokenizer. validateNextToken ( );
			if ( sqlTokenizer. peekAtNextToken ( ). equals ( "." ) )
			{
				// strToken, in fact, contains table alias and column name should follow
				strRightTableAlias = strToken;
				validateTableAlias ( strRightTableAlias );
				sqlTokenizer. getNextToken ( ); // consume period '.'
				
				// Extract column name
				sqlTokenizer. validateNextToken ( );
				strRightColumnName = sqlTokenizer. getNextToken ( );
				validateColumnName ( strRightColumnName );
			}
			else
				// strToken, in fact, contains column name and '=' should follow
				strRightColumnName = strToken;
			
			sqlcMDRight = new SQLColumnMetaData (
										SQLTools. removeEnclosers ( strRightColumnName, '[', ']' ) );
			if ( strRightTableAlias != null )
				sqlcMDRight. setTableAlias (
									SQLTools. removeEnclosers ( strRightTableAlias, '[', ']' ) );
				
			sqltMDJoin. addJOINCondition (
							new JoinCondition ( sqlcMDLeft, sqlcMDRight ) );

			/*
				Now we can have :
					the end of query,
					AND keyword,
					LEFT keyword,
					some other sql keyword
			*/
			if ( !sqlTokenizer. hasMoreTokens ( ) )
			{
				sqlCommand. addJOINTable ( sqltMDJoin );
				
				return;
			}
			
			strToken = sqlTokenizer. getNextToken ( );
			if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_AND ) )
			{
				// Multi-column join
				sqlTokenizer. validateNextToken ( );
				
				continue;
			}
			else if ( strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_LEFT ) )
			{
				// Multi-table join
				sqlCommand. addJOINTable ( sqltMDJoin );	
				parseJOIN ( sqlTokenizer, sqlCommand );
				
				return;
			}
			else
			{
				sqlCommand. addJOINTable ( sqltMDJoin );	
				sqlTokenizer. rollBack ( 1 );
				
				return;
			}
		} // end of while ( sqlTokenizer. hasMoreTokens ( ) )
	} // end of private static void parseJOIN ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	
	/**
	*	Parses DELETE part of DELETE clause. Does NOT parse WHERE part.
	*
	*	sqlTokenizer already consumed DELETE keyword and points right after it.
	*
	*	Supported DELETE syntax:
	*		DELETE [*] FROM TableName [AS] TableAlias
	*		WHERE[4D] [TableName|TableAlias.]ColumnName   =|<>|<|>|<=|>=   Number|String|Date   AND|OR   ...
	*/
	private static void parseDELETE ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	{
		sqlTokenizer. validateNextToken ( );

		// Lets see if there is an optional *
		if ( sqlTokenizer. peekAtNextToken ( ). equals ( "*" ) )
		{
			sqlTokenizer. getNextToken ( ); // skip *
			sqlTokenizer. validateNextToken ( ); // FROM keyword should follow
		}
		
		// Now strToken must contain FROM keyword
		sqlTokenizer. validateNextToken ( SQLCommand. SQL_KEYWORD_FROM );
		sqlTokenizer. getNextToken ( ); // skip FROM keyword
		
		// Table name should follow
		sqlTokenizer. validateNextToken ( );
		String		strToken = sqlTokenizer. getNextToken ( );
		validateColumnName ( strToken );
		SQLTableMetaData			sqltMDFROM = new SQLTableMetaData (
															SQLTools. removeEnclosers ( strToken, '[', ']' ) );
		
		if ( !sqlTokenizer. hasMoreTokens ( ) ) // Simple 'DELETE FROM TableName' command
		{
			sqlCommand. addFROMTable ( sqltMDFROM );
			return;
		}
		
		// Now I can have either AS or WHERE or WHERE4D
		
		if (	sqlTokenizer. peekAtNextToken ( ). toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_WHERE ) ||
			sqlTokenizer. peekAtNextToken ( ). toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_WHERE4D ) )
		{
			sqlCommand. addFROMTable ( sqltMDFROM );
			return;
		}
		
		// Now the only possible token is AS
		sqlTokenizer. validateNextToken ( SQLCommand. SQL_KEYWORD_AS );
		sqlTokenizer. getNextToken ( ); // skip AS keyword
		sqlTokenizer. validateNextToken ( ); // Table alias should follow
		
		// Deal with table alias
		strToken = sqlTokenizer. getNextToken ( );
		validateTableAlias ( strToken );
		sqltMDFROM. setAlias ( strToken );
		sqlCommand. addFROMTable ( sqltMDFROM );
	} // end of private static void parseDELETE ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	
	/**
	*	Parses INSERT INTO <TableName> part of the DELETE command.
	*	Does NOT parse anything else.
	*
	*	Input parameter sqlTokenizer already consumed INSERT keyword.
	*
	*	Supported syntax:
	*		INSERT INTO <TableName> [( ColumnName1, ... )]
	*		VALUES ( ColumnValue1, ... )
	*/
	private static void parseINSERT (
							SQLTokenizer sqlTokenizer,
							SQLCommand sqlCommand,
							int [ ] sznParameterIndex ) throws Exception
	{
		// Make sure that INTO keyword is present
		sqlTokenizer. validateNextToken ( SQLCommand. SQL_KEYWORD_INTO );
		sqlTokenizer. getNextToken ( ); // skip INTO keyword
		
		// Table name should follow
		sqlTokenizer. validateNextToken ( );
		String			strToken = sqlTokenizer. getNextToken ( );
		validateTableName ( strToken );
		SQLTableMetaData			sqltMDINTO = new SQLTableMetaData (
															SQLTools. removeEnclosers ( strToken, '[', ']' ) );
		
		// INTO table is stored as FROM table
		sqlCommand. addFROMTable ( sqltMDINTO );
		
		// A '(' or 'VALUES' should follow
		sqlTokenizer. validateNextToken ( );
		if ( sqlTokenizer. peekAtNextToken ( ). equals ( "(" ) )
			parseINSERTColumnList ( sqlTokenizer, sqlCommand );
			
		// The only alternative is 'VALUES'
		sqlTokenizer. validateNextToken ( SQLCommand. SQL_KEYWORD_VALUES );
		sqlTokenizer. getNextToken ( ); // skip 'VALUES'
		
		parseINSERTValueList ( sqlTokenizer, sqlCommand, sznParameterIndex );
	}
	
	/**
	*	Parses column name list for INSERT command:
	*		( ColumnName1, ... )
	*
	*	The list can not be empty.
	*	Input parameter sqlTokenizer points right before '(', but I will double-check
	*	this anyway.
	*/
	private static void parseINSERTColumnList ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	{
		sqlTokenizer. validateNextToken ( "(" );
		sqlTokenizer. getNextToken ( ); // skip '('
		
		// Parse column name list
		String			strToken = null;
		while ( true )
		{
			// In one cycle of this loop I parse exactly one column name
			sqlTokenizer. validateNextToken ( );
			strToken = sqlTokenizer. getNextToken ( );
			validateColumnName ( strToken );
			
			Value		val = new Value ( );
			val. setColumnName ( SQLTools. removeEnclosers ( strToken, '[', ']' ) );
			sqlCommand. addValue ( val );
			
			// Now I can either have a closing ')' or a ','
			sqlTokenizer. validateNextToken ( );
			if ( sqlTokenizer. peekAtNextToken ( ). equals ( ")" ) )
			{
				sqlTokenizer. getNextToken ( ); // skip ')'
				return; // I'm done here
			}
				
			sqlTokenizer. validateNextToken ( "," );
			sqlTokenizer. getNextToken ( ); // skip ','
		}
	} // end of private static void parseINSERTColumnList ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	
	/**
	*	Parses column value list for INSERT command:
	*		( ColumnValue1, ... )
	*
	*	The list can not be empty.
	*	Input parameter sqlTokenizer points right before '(', but I will double-check
	*	this anyway.
	*/
	private static void parseINSERTValueList (
								SQLTokenizer sqlTokenizer,
								SQLCommand sqlCommand,
								int [ ] sznParameterIndex ) throws Exception
	{
		sqlTokenizer. validateNextToken ( "(" );
		sqlTokenizer. getNextToken ( ); // skip '('
		
		// Parse column name list
		String			strToken = null;
		String			strValue = null;
		Parameter		param = null;
		int				nIndex = 0;
		while ( true )
		{
			// In one cycle of this loop I parse exactly one column value
			sqlTokenizer. validateNextToken ( );
			strToken = sqlTokenizer. getNextToken ( );
			
			strValue = null;
			param = null;
			if ( strToken. equals ( "?" ) )
			{
				sqlCommand. addParameter ( null );
				param = new Parameter ( sznParameterIndex [ 0 ] );
				sznParameterIndex [ 0 ] ++;
			}
			else if ( SQLTools. isString ( strToken ) )
				strValue = SQLTools. removeEnclosers ( strToken, '\'', '\'' );
			else if (	strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_TRUE ) ||
						strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_FALSE ) )
				strValue = strToken. toUpperCase ( );
			else // Value is a number
				// Complete number is read in the SQLTokenizer
				strValue = strToken;
				
			if ( strValue != null )
				sqlCommand. setValue ( nIndex, strValue );
			else if ( param != null )
				sqlCommand. setValue ( nIndex, param );
			else
				throw new Exception (
									"com.fourd.jdbc.SQLParser::parseINSERTValueList: " +
									"Unexpected value - " + strToken );
			
			// Now I can either have a closing ')' or a ','
			sqlTokenizer. validateNextToken ( );
			if ( sqlTokenizer. peekAtNextToken ( ). equals ( ")" ) )
			{
				sqlTokenizer. getNextToken ( ); // skip ')'
				return; // I'm done here
			}
				
			sqlTokenizer. validateNextToken ( "," );
			sqlTokenizer. getNextToken ( ); // skip ','
			nIndex++;
		}
	} // end of private static void parseINSERTValueList ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	
	/**
	*	Parses UPDATE <TableName> part of the UPDATE command.
	*	Does NOT parse anything else.
	*
	*	Input parameter sqlTokenizer already consumed UPDATE keyword.
	*
	*	Supported syntax:
	*		UPDATE <TableName>
	*		SET ColumnName1 = ColumnValue1, ...
	*		WHERE[4D] ...
	*/
	private static void parseUPDATE ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	{
		// Table name should follow
		sqlTokenizer. validateNextToken ( );
		String			strToken = sqlTokenizer. getNextToken ( );
		validateTableName ( strToken );
		SQLTableMetaData		sqltMDUPDATE = new SQLTableMetaData (
														SQLTools. removeEnclosers (
																		strToken, '[', ']' ) );
		
		// UPDATE table is stored as FROM table
		sqlCommand. addFROMTable ( sqltMDUPDATE );
	} // end of private static void parseUPDATE ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
		
	/**
	*	Parses SET ColumnName1 = ColumnValue1, ...
	*/
	private static void parseUPDATEList (
								SQLTokenizer sqlTokenizer,
								SQLCommand sqlCommand,
								int [ ] sznParameterIndex ) throws Exception
	{
		// SET keyword should be present
		sqlTokenizer. validateNextToken ( SQLCommand. SQL_KEYWORD_SET );
		sqlTokenizer. getNextToken ( ); // skip SET keyword
		
		// Parse ColumnName = ColumnValues pairs
		String			strToken = null;
		Value			val = null;
		Parameter		param = null;
		String			strValue = null;
		while ( true )
		{
			// Column name
			sqlTokenizer. validateNextToken ( );
			strToken = sqlTokenizer. getNextToken ( );
			validateColumnName ( strToken );
			val = new Value ( );
			val. setColumnName ( strToken );
			
			sqlTokenizer. validateNextToken ( "=" );
			sqlTokenizer. getNextToken ( ); // skip '='
			
			// Column value
			sqlTokenizer. validateNextToken ( );
			strToken = sqlTokenizer. getNextToken ( );
			
			strValue = null;
			param = null;
			if ( strToken. equals ( "?" ) )
			{
				sqlCommand. addParameter ( null );
				param = new Parameter ( sznParameterIndex [ 0 ] );
				sznParameterIndex [ 0 ] ++;
			}
			else if ( SQLTools. isString ( strToken ) )
				strValue = SQLTools. removeEnclosers ( strToken, '\'', '\'' );
			else if (	strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_TRUE ) ||
						strToken. toUpperCase ( ). equals ( SQLCommand. SQL_KEYWORD_FALSE ) )
				strValue = strToken. toUpperCase ( );
			/*else // Value is a number
				// Complete number is read in the SQLTokenizer
				; //strValue = strToken; */
				
			if ( strValue != null )
				val. setValue ( strValue );
			else if ( param != null )
				val. setParameter ( param );
			else
				throw new Exception (
									"com.fourd.jdbc.SQLParser::parseUPDATEList: " +
									"Unexpected value - " + strToken );
			
			sqlCommand. addValue ( val );
				
			// Now I can have either the end of query, a ',' or 'WHERE'
			if ( !sqlTokenizer. hasMoreTokens ( ) )
				return;
				
			if ( !sqlTokenizer. peekAtNextToken ( ). equals ( "," ) )
				// Not the end of query and not ','.
				return;
			
			sqlTokenizer. getNextToken ( ); // skip ','
		} // end of while ( true )
	} // end of private static void parseUPDATEList ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	
	/**
	*	Parses stored procedure call. The format is :
	*		{ CALL <stored procedure name> ( [ ?, ... ] ) }
	*
	*	Input parameter sqlTokenizer already consumed the opening {
	*/
	private static void parseCALL ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception
	{
		sqlTokenizer. validateNextToken ( SQLCommand. SQL_KEYWORD_CALL );
		sqlTokenizer. getNextToken ( ); // skip CALL
		
		sqlTokenizer. validateNextToken ( );
		String			strToken = sqlTokenizer. getNextToken ( ); // stored procedure name
		validateStoredProcedureName ( strToken );
		sqlCommand. setStoredProcedureName (
						SQLTools. removeEnclosers ( strToken, '[', ']' ) );
		
		sqlTokenizer. validateNextToken ( "(" );
		sqlTokenizer. getNextToken ( ); // skip '('
		
		// Lets see if this is the case of '( )' - no parameters
		sqlTokenizer. validateNextToken ( );
		if ( sqlTokenizer. peekAtNextToken ( ). equals ( ")" ) )
		{
			sqlTokenizer. getNextToken ( ); // skip )
		}
		else
		{
			// Parse a list of "? ,"
			while ( true )
			{
				// I expect only '?' at this point
				sqlTokenizer. validateNextToken ( "?" );
				sqlTokenizer. getNextToken ( ); // skip '?'
				sqlCommand. addParameter ( null );
				
				// Now there can be a ',' or ')'
				sqlTokenizer. validateNextToken ( );
				strToken = sqlTokenizer. peekAtNextToken ( );
				if ( strToken. equals ( "," ) ) // more '?,' will follow
				{
					sqlTokenizer. getNextToken ( ); // skip ','
				
					continue;
				}
			
				// Now the only choice is ')'
				sqlTokenizer. validateNextToken ( ")" );
				sqlTokenizer. getNextToken ( ); // skip ')'
			
				break;
			} // end of while ( true )
		} // end of else
		
		// There should be a closing '}'
		sqlTokenizer. validateNextToken ( "}" );
		sqlTokenizer. getNextToken ( ); // skip '}'
	} // end of private static void parseCALL ( SQLTokenizer sqlTokenizer, SQLCommand sqlCommand ) throws Exception

	/**
	*	Acts as an assert macros. Does nothing if strColumnName represents a valid column name.
	*	Throws an exception otherwise.
	*/
	private static void validateColumnName ( String strColumnName ) throws Exception
	{
		// Lets make sure that column name is valid
		if ( SQLCommand. isSQLKeyword ( strColumnName ) )
			throw new Exception (
							"com.fourd.jdbc.SQLParser.validateColumnName: Column name " +
							strColumnName +
							" is an SQL keyword. It should be escaped with [ and ]" );
		if ( !SQLTools. isValidColumnName ( strColumnName ) )
			throw new Exception (
							"com.fourd.jdbc.SQLParser.validateColumnName: Invalid column name - " +
							strColumnName );
	}

	/**
	*	Acts as an assert macros. Does nothing if strTableName represents a valid table name.
	*	Throws an exception otherwise.
	*/
	private static void validateTableName ( String strTableName ) throws Exception
	{
		// Lets make sure that table name is valid
		if ( SQLCommand. isSQLKeyword ( strTableName ) )
			throw new Exception (
							"com.fourd.jdbc.SQLParser.validateTableName: Table name " +
							strTableName +
							" is an SQL keyword. It should be escaped with [ and ]" );
		if ( !SQLTools. isValidTableName ( strTableName ) )
			throw new Exception (
							"com.fourd.jdbc.SQLParser.validateTableName: Invalid table name - " +
							strTableName );
	}
	
	/**
	*	Acts as an assert macros. Does nothing if strSName represents a valid 
	*	stored procedure name. Throws an exception otherwise.
	*/
	private static void validateStoredProcedureName ( String strSPName ) throws Exception
	{
		// Lets make sure that stored procedure name is valid
		if ( SQLCommand. isSQLKeyword ( strSPName ) )
			throw new Exception (
							"com.fourd.jdbc.SQLParser.validateStoredProcedureName: " +
							"Column name " + strSPName +
							" is an SQL keyword. It should be escaped with [ and ]" );
		
		// TODO: implement proper validation
		if ( !SQLTools. isValidColumnName ( strSPName ) )
			throw new Exception (
							"com.fourd.jdbc.SQLParser.validateStoredProcedureName: " +
							"Invalid stored procedure name - " + strSPName );
	} // end of private static void validateStoredProcedureName ( String strSPName ) throws Exception
	
	/**
	*	Acts as an assert macros. Does nothing if strTableAlias represents a valid table alias.
	*	Throws an exception otherwise.
	*/
	private static void validateTableAlias ( String strTableAlias ) throws Exception
	{
		if ( !SQLTools. isValidTableAlias ( strTableAlias ) )
			throw new Exception (
							"com.fourd.jdbc.SQLParser.validateTableAlias: " +
							"Invalid table alias - " + strTableAlias );
	}
}