/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.sql.*;
import java.util.*;

import j4d.open.*;

/**
*	Stores information about table.
*	Theoretically this class does not need any real functionality.
*/
class SQLTableMetaData
{
	// Stores real name of the table in the database.
	private String						m_strDBName = null;
		
	/** 
	*	Stores alias name of the table in the query.
	*	Alias of the table name comes from
	*		SELECT DBT1.DBColumnName AS ColumnAlias
	*		FROM DBTableName AS DBT1
	*
	*	In the example above, DBT1 is an alias.
	*/
	private String						m_strAlias = null;
	
	// Is used to reference a table in 4D Open for java
	private short						m_nNumber = -1;
	
	private Vector						m_vctrColumns = new Vector ( );
	private Vector						m_vctrJOINConditions = new Vector ( );
	
	/**
	*	Instances of this class may be used to represent data base meta data or
	*	meta data produced during query parsing.
	*	In case of data base meta data, column information may be fetched in "lazy"
	*	way. That is, it is fetched only when requested. In order to do this, 
	*	the following class members are used.
	*/
	// DataBaseMetaData that generated this instance
	private DatabaseMetaDataImpl						m_dbMetaDataImpl = null;
	// opTable that represents table which meta data contains in the instance of this class
	private opTable								m_opTable = null;
	// Boolean flag that indicates whether or not column's meta data is already pre-loaded
	private boolean								m_bColumnsMDLoaded = false;
	
	SQLTableMetaData ( )
	{
		;
	}
	
	SQLTableMetaData ( String strDBName ) throws IllegalArgumentException
	{
		setDBName ( strDBName );
	}
	
	SQLTableMetaData (
					DatabaseMetaDataImpl dbMetaDataImpl,
					opTable opt ) throws IllegalArgumentException
	{
		if ( dbMetaDataImpl == null || opt == null )
			throw new IllegalArgumentException (
						"com.fourd.jdbc.SQLTableMetaData::SQLTableMetaData: " +
						"Input parameters can not be null" );
						
		m_dbMetaDataImpl = dbMetaDataImpl;
		m_opTable = opt;
	}
	
	void setDBName ( String strDBName ) throws IllegalArgumentException
	{
		if ( strDBName == null )
			throw new IllegalArgumentException (
						"com.fourd.jdbc.SQLTableMetaData.setDBName: Table name can not be null" );
						
		m_strDBName = strDBName;
	}
	
	String getDBName ( )
	{
		return m_strDBName;
	}
	
	/**
	*	Sets the alias of the table in the query.
	*/
	void setAlias ( String strAlias ) throws IllegalArgumentException
	{
		if ( strAlias == null )
			throw new IllegalArgumentException (
						"com.fourd.jdbc.SQLTableMetaData.setAlias: Table alias can not be null" );
						
		m_strAlias = strAlias;
	}
	
	String getAlias ( )
	{
		return m_strAlias;
	}
	
	short getNumber ( )
	{
		return m_nNumber;
	}
	
	void setNumber ( short nNumber )
	{
		m_nNumber = nNumber;
	}
	
	void addColumn ( SQLColumnMetaData sqlcMD ) throws SQLException
	{
		if ( sqlcMD == null )
			throw new SQLException ( "com.fourd.jdbc.SQLTableMetaData::addColumn: Column must not be null" );
			
		m_vctrColumns. addElement ( sqlcMD );
	}

	void addColumns ( SQLColumnMetaData [ ] szsqlcMD ) throws SQLException
	{
		if ( szsqlcMD == null )
			throw new SQLException ( "com.fourd.jdbc.SQLTableMetaData::addColumns: Columns must not be null" );
		
		for ( int i = 0; i < szsqlcMD. length; i++ )
			addColumn ( szsqlcMD [ i ] );
	}
	
	int getColumnCount ( ) throws SQLException
	{
		loadColumnsMetaData ( );
		
		return m_vctrColumns. size ( );
	}
	
	SQLColumnMetaData getColumnMetaData ( String strColumnName ) throws SQLException
	{
		loadColumnsMetaData ( );
		
		if ( strColumnName == null )
			throw new SQLException ( "com.fourd.jdbc.SQLTableMetaData::getColumn: Column name must not be null" );
			
		SQLColumnMetaData		sqlcMD = null;
		for ( int i = 0; i < m_vctrColumns. size ( ); i++ )
		{
			sqlcMD = ( SQLColumnMetaData ) m_vctrColumns. elementAt ( i );
			if ( sqlcMD. getDBName ( ). toUpperCase ( ). equals ( strColumnName. toUpperCase ( ) ) )
				return sqlcMD;
		}
		
		throw new SQLException ( "com.fourd.jdbc.SQLTableMetaData::getColumn: Unknown column name " + strColumnName );
	}

	SQLColumnMetaData [ ] getColumns ( ) throws SQLException
	{
		loadColumnsMetaData ( );
		
		return ( SQLColumnMetaData [ ] ) m_vctrColumns. toArray ( new SQLColumnMetaData [ m_vctrColumns. size ( ) ] );
	}
	
	/**
	*	Returns "field number" of a given column in the table in the DB.
	*
	*	Throws SQLException if anything goes wrong.
	*/
	int getColumnNumber ( String strColumnName ) throws SQLException
	{
		loadColumnsMetaData ( );
		
		if ( strColumnName == null )
			throw new SQLException ( "com.fourd.jdbc.SQLTableMetaData::getColumnIndex: Column name must not be null" );
			
		SQLColumnMetaData		sqlcMD = null;
		for ( int i = 0; i < m_vctrColumns. size ( ); i++ )
		{
			sqlcMD = ( SQLColumnMetaData ) m_vctrColumns. elementAt ( i );
			if ( sqlcMD. getDBName ( ). toUpperCase ( ). equals ( strColumnName. toUpperCase ( ) ) )
				return sqlcMD. getNumber ( );
		}
		
		throw new SQLException ( "com.fourd.jdbc.SQLTableMetaData::getColumnIndex: Unknown column name " + strColumnName );
	}

	boolean columnExists ( String strColumnName ) throws SQLException
	{
		loadColumnsMetaData ( );
		
		if ( strColumnName. equals ( "*" ) )
			return true;
			
		SQLColumnMetaData		sqlcMD = null;
		for ( int i = 0; i < m_vctrColumns. size ( ); i++ )
		{
			sqlcMD = ( SQLColumnMetaData ) m_vctrColumns. elementAt ( i );
			if (
				sqlcMD. getDBName ( ) != null &&
				sqlcMD. getDBName ( ). toUpperCase ( ). equals ( strColumnName. toUpperCase ( ) ) )
				return true;
		}
		
		return false;
	}
	
	void addJOINCondition ( JoinCondition jc ) throws SQLException
	{
		if ( jc == null )
			throw new SQLException (
								"com.fourd.jdbc.SQLTableMetaData::addJOINCondition: " +
								"JOIN condition can not be null" );
			
		m_vctrJOINConditions. addElement ( jc );
	}

	JoinCondition [ ] getJOINConditions ( )
	{
		return ( JoinCondition [ ] ) m_vctrJOINConditions. toArray (
									new JoinCondition [ m_vctrJOINConditions. size ( ) ] );
	}
	
	/**
	*	Returns the name or the alias of the table that we are JOINed to.
	*	Throws an exception if we are not JOINed to anybody or 
	*	if anything else goes wrong.
	*/
	String getTableAliasJoinedTo ( String strFROMTableName ) throws SQLException
	{
		JoinCondition [ ]			szjConditions = getJOINConditions ( );
		if ( szjConditions. length == 0 )
			throw new SQLException (
								"com.fourd.jdbc.SQLTableMetaData::getTableNameJoinedTo: " +
								"Table " + getDBName ( ) + " is not joined" );
								
		SQLColumnMetaData		sqlcMDLeft = szjConditions [ 0 ]. getLeftColumn ( );
		if ( sqlcMDLeft. getTableAlias ( ) == null )
			// Just a column name without a table name/alias and thus this column name is from the single FROM table
			return strFROMTableName;
		else
			return sqlcMDLeft. getTableAlias ( );
	} // end of String getTableAliasJoinedTo ( ) throws SQLException
	
	private void loadColumnsMetaData ( ) throws SQLException
	{
		if ( m_bColumnsMDLoaded || m_dbMetaDataImpl == null )
			return;
	
//System. out. println ( "Lazy loading meta data for table " + m_strDBName );
		addColumns ( m_dbMetaDataImpl. lazyLoadColumnsMetaData ( m_opTable ) );
		m_bColumnsMDLoaded = true;
	}
}