/* ====================================================================
  * Copyright (c) 2004 4D, SA.  All rights
  * reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
  *
  * 1. Redistributions of source code must retain the above copyright
  *    notice, this list of conditions and the following disclaimer.
  *
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice.
  *
  * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
  * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED.  IN NO EVENT SHALL 4D, SA ITS SUBSIDIARIES OR
  * ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
  * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
  * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
  * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  * ====================================================================
  */

package com.fourd.jdbc;

import java.sql.*;
import java.text.*;
import java.util.*;
import java.io.*;
import java.math.*;

import j4d.open.*;

/**
*	Contains various static helper methods.
*/
class SQLTools
{
	/**
	*	Removes enclosing characters. Does this only if both - opening and closing characters are present.
	*	Returns original string unchanged otherwise.
	*/
	static String removeEnclosers ( String strData, char chOpen, char chClose )
	{
		if ( strData == null )
			return null;
			
		if ( strData. charAt ( 0 ) != chOpen || strData. charAt ( strData. length ( ) - 1 ) != chClose )
			return strData;
			
		return strData. substring ( 1, strData. length ( ) - 1 );
	}
	
	/**
	*	Determines if strTableName represents a valid 4D table name.
	*	Table name may be enclosed in square brackets.
	*	Returns true if the table name is valid, false - otherwise.
	*/
	static boolean isValidTableName ( String strTableName )
	{	
		// TODO: find out what are the restrictions for 4D table names
		
		return SQLColumnMetaData. isValidColumnName ( strTableName );
	}

	static boolean isValidTableAlias ( String strTableAlias )
	{
		return SQLColumnMetaData. isValidColumnName ( strTableAlias );
	}
	
	static boolean isValidColumnName ( String strColumnName )
	{
		return SQLColumnMetaData. isValidColumnName ( strColumnName );
	}

	/**
	*	Column alias is anything enclosed in double quotes.
	*/
	static boolean isValidColumnAlias ( String strColumnAlias )
	{
		if ( strColumnAlias == null || strColumnAlias. length ( ) < 2 )
			return false;
			
		return strColumnAlias. charAt ( 0 ) == '"' && strColumnAlias. charAt ( strColumnAlias. length ( ) - 1 ) == '"';
	}
	
	/**
	*	Determines if str represents a valid string value. Valid string value should be enclosed in 
	*	single quotes. Single quote inside the string should be escaped with another single quote.
	*	That is, single quotes can be present only in pairs inside a valid string.
	*/
	static boolean isValidString ( String str )
	{
		if ( str == null || str. length ( ) < 2 )
			return false;
			
		if ( str. charAt ( 0 ) != '\'' && str. charAt ( str. length ( ) - 1 ) != '\'' )
			return false;
			
		// TODO: make sure that single quotes inside the string are properly escaped
		
		return true;
	}

	/**
	*	Determines if str represents a single-quoted string.
	*/
	static boolean isString ( String str )
	{
		if ( str == null || str. length ( ) < 2 )
			return false;
			
		if ( str. charAt ( 0 ) != '\'' && str. charAt ( str. length ( ) - 1 ) != '\'' )
			return false;
			
		return true;
	}
	
	/**
	*	TODO: this is a complite duplicate of the RSDatCache:convertToJavaObject ( opData )
	*	Should have only one impementation.
	*/
	static Object convertToJavaObject ( opData opD ) throws SQLException
	{
		if ( opD == null )
			return null;
			
		// TODO: possible location for memory optimization
		if ( opD. mType == opConstants. ALPHANUMERIC )
			return new String ( opD. mString );
		else if ( opD. mType == opConstants. BOOLEAN )
			return new Boolean ( opD. mBoolean );
		else if ( opD. mType == opConstants. INTEGER )
			return new Short ( opD. mInteger );
		else if ( opD. mType == opConstants. LONGINTEGER )
			return new Integer ( opD. mLongInteger );
		else if ( opD. mType == opConstants. REAL )
			return new Double ( opD. mReal );
		else if ( opD. mType == opConstants. TEXT )
			return new String ( opD. mString );
		else if ( opD. mType == opConstants. TIME )
			return new java.sql.Time ( opD. mTime. getTime ( ). getTime ( ) );
		else if ( opD. mType == opConstants. DATE )
		{
			if ( opD. mDate == null )
				return null;
			
			return new java.sql.Date ( opD. mDate. getTime ( ). getTime ( ) );
		}
		else if ( opD. mType == opConstants. BLOB )
			if ( opD. mBlob. length == opD. mBlobSize )
				return opD. mBlob;
			else
			{
				byte [ ]		szbtResult = new byte [ opD. mBlobSize ];
				System. arraycopy (
								opD. mBlob, 0,
								szbtResult, 0,
								opD. mBlobSize );
									
				return szbtResult;
			}
		else if ( opD. mType == opConstants. PICTURE )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convertToJavaObject: " +
								"Unimplemented data type - PICTURE" );
		else if ( opD. mType == opConstants. SUBTABLE )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convertToJavaObject: " +
								"Unimplemented data type - SUBTABLE" );
		else if ( opD. mType == opConstants. ALPHANUMERIC2 )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convertToJavaObject: " +
								"Unimplemented data type - ALPHANUMERIC2" );
			
		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convertToJavaObject: " +
							"Unknown data type - " + opD. mType );
	}

	static java.util.Date parseDate ( String strDate )
	{
		if ( strDate == null )
			return null;
			
		String [ ]					szDateFormats = new String [ ] {
															"EEE, dd MMM yyyy hh:mm:ss z",
															"dd MMM yyyy hh:mm:ss z",
															"EEE, dd MMM yyyy hh:mm:ss",
															"dd MMM yyyy hh:mm:ss",
															"dd MMM yyyy" };
		java.util.Date		dtResult = null;
		SimpleDateFormat	sdFormat = null;
		for ( int i = 0; i < szDateFormats. length; i++ )
			try
			{
				sdFormat = new SimpleDateFormat ( szDateFormats [ i ] );
				dtResult = sdFormat. parse ( strDate );
				
				return dtResult;
			}
			catch ( ParseException pE )
			{
				;
			}
			
		// Unsupported date format or invalid date
		return null;
	}
	
	/**
	*	Creates opData object from java object.
	*	Tries not only to map directly java type x to corresponding 4D type x but
	*	also to convert types if necessary. For example, let say input object parameter
	*	is of java type Integer and nopType is 4d alfanumeric type. In this case
	*	method will try to convert java Integer to 4D alfanumeric by converting
	*	java Integer to its string representation.
	*/
	static opData createopData ( Object objValue, int nopType ) throws SQLException
	{
		/*
		*	objValue can be instance of one of the following:
		*
		*	InputStream
		*	BigDecimal
		*	Boolean
		*	Byte
		*	byte [ ]
		*	Reader
		*	java.sql.Date
		*	Double
		*	Float
		*	Integer
		*	Long
		*	Object
		*	Short
		*	String
		*	java.sql.Time
		*	java.sql.Timestamp
		*/
		Object		objCast = null;
		if ( nopType == opConstants. ALPHANUMERIC )
		{
			objCast = convert ( objValue, String. class );
			
			return new opData ( nopType, ( String ) objCast );
		}
		else if ( nopType == opConstants. BOOLEAN )
		{
			objCast = convert ( objValue, Boolean. class );
			
			return new opData ( nopType, ( ( Boolean ) objCast ). booleanValue ( ) );
		}
		else if ( nopType == opConstants. INTEGER )
		{
			objCast = convert ( objValue, Short. class );
			
			return new opData ( nopType, ( ( Short ) objCast ). shortValue ( ) );
		}
		else if ( nopType == opConstants. LONGINTEGER )
		{
			objCast = convert ( objValue, Integer. class );
			
			return new opData ( nopType, ( ( Integer ) objCast ). intValue ( ) );
		}
		else if ( nopType == opConstants. REAL )
		{
			objCast = convert ( objValue, Double. class );
			
			return new opData ( nopType, ( ( Double ) objCast ). doubleValue ( ) );
		}
		else if ( nopType == opConstants. TEXT )
		{
			objCast = convert ( objValue, String. class );
			
			return new opData ( nopType, ( String ) objCast );
		}
		else if ( nopType == opConstants. DATE )
		{
			objCast = convert ( objValue, java.sql.Date. class );
			GregorianCalendar		gCalendar = new GregorianCalendar ( );
			gCalendar. setTime ( ( java.sql.Date ) objCast );
			
			return new opData ( nopType, gCalendar );
		}
		else if ( nopType == opConstants. TIME )
		{
			objCast = convert ( objValue, Time. class );
			GregorianCalendar		gCalendar = new GregorianCalendar ( );
			gCalendar. setTime ( ( Time ) objCast );
			
			return new opData ( nopType, gCalendar );
		}
		else if ( nopType == opConstants. BLOB )
		{
			objCast = convert ( objValue, byte [ ]. class );
			byte [ ]	szbtValue = ( byte [ ] ) objCast;
			opData		opdResult = new opData ( nopType, szbtValue );
			opdResult. mBlobSize = szbtValue. length;
			
			return opdResult;
		}
		/*else if ( nopType == opConstants. PICTURE )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::convertToJavaObject: " +
								"Unimplemented data type - PICTURE" );
		else if ( nopType == opConstants. SUBTABLE )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::convertToJavaObject: " +
								"Unimplemented data type - SUBTABLE" );
		else if ( nopType == opConstants. ALPHANUMERIC2 )
			throw new SQLException (
								"com.fourd.jdbc.RSDataCache::convertToJavaObject: " +
								"Unimplemented data type - ALPHANUMERIC2" );
		*/
		
		throw new SQLException (
							"com.fourd.jdbc.SQLTools::createopData: " +
							"Unimplemented 4D data type - " + nopType );
	}
	
	/**
	*	Tries to concert objSource to instance of class clsDestination.
	*	If convertion is successfull then result is returned as instance of Object,
	*	otherwise an SQLException is thrown.
	*/
	static Object convert ( Object objSource, Class clsDestination ) throws SQLException
	{
		if ( objSource == null )
			return null;
			
		if ( objSource instanceof Boolean )
			return convert ( ( Boolean ) objSource, clsDestination );
		else if ( objSource instanceof Byte )
			return convert ( ( Byte ) objSource, clsDestination );
		else if ( objSource instanceof byte [ ] )
			return convert ( ( byte [ ] ) objSource, clsDestination );
		else if ( objSource instanceof java.sql.Date )
			return convert ( ( java.sql.Date ) objSource, clsDestination );
		else if ( objSource instanceof Double )
			return convert ( ( Double ) objSource, clsDestination );
		else if ( objSource instanceof Float )
			return convert ( ( Float ) objSource, clsDestination );
		else if ( objSource instanceof Integer )
			return convert ( ( Integer ) objSource, clsDestination );
		else if ( objSource instanceof Long )
			return convert ( ( Long ) objSource, clsDestination );
		else if ( objSource instanceof Short )
			return convert ( ( Short ) objSource, clsDestination );
		else if ( objSource instanceof String )
			return convert ( ( String ) objSource, clsDestination );
		else if ( objSource instanceof Time )
			return convert ( ( Time ) objSource, clsDestination );
		
		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
							"instance of " + objSource. getClass ( ). getName ( ) +
							" to " + clsDestination. getName ( ) );
	} // end of static Object convert ( Object objSource, Class clsDestination ) throws SQLException
	
	static Object convert ( String strSource, Class clsDestination ) throws SQLException
	{
		if ( strSource == null )
			return null;
			
		if ( clsDestination == InputStream. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"String to InputStream" );
		else if ( clsDestination == BigDecimal. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"String to BigDecimal" );
		else if ( clsDestination == Boolean. class )
		{
			if ( strSource. toLowerCase ( ). equals ( "true" ) || strSource. equals ( "1" ) )
				return new Boolean ( true );
			else if ( strSource. toLowerCase ( ). equals ( "false" ) || strSource. equals ( "0" ) )
				return new Boolean ( false );
			else
				throw new SQLException (
									"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
									"string \"" + strSource + "\" to Boolean" );
		}
		else if ( clsDestination == Byte. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"String to Byte" );
		else if ( clsDestination == byte [ ]. class )
			return strSource. getBytes ( );
		else if ( clsDestination == Reader. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"String to Reader" );
		else if ( clsDestination == java.sql.Date. class )
		{
			String [ ]					szDateFormats = new String [ ] {
																"EEE, dd MMM yyyy hh:mm:ss z",
																"dd MMM yyyy hh:mm:ss z",
																"EEE, dd MMM yyyy hh:mm:ss",
																"dd MMM yyyy hh:mm:ss",
																"dd MMM yyyy" };
			java.util.Date				dtResult = null;
			for ( int i = 0; i < szDateFormats. length; i++ )
			{
				SimpleDateFormat			sdFormat = new SimpleDateFormat ( szDateFormats [ i ] );
				try
				{
					dtResult = sdFormat. parse ( strSource );
					return new java.sql.Date ( dtResult. getTime ( ) );
				}
				catch ( ParseException pE ) { ; }
			}
			
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"String " + strSource + " to Date" );
		}
		
		try
		{
			if ( clsDestination == Double. class )
				return new Double ( strSource );
			else if ( clsDestination == Float. class )
				return new Float ( strSource );
			else if ( clsDestination == Integer. class )
				return new Integer ( strSource );
			else if ( clsDestination == Long. class )
				return new Long ( strSource );
			else if ( clsDestination == Short. class )
				return new Short ( strSource );
		}
		catch ( NumberFormatException nfE )
		{
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: " + strSource +
								"is not a valid number" );
		}
		
		if ( clsDestination == Object. class )
			return strSource;
		else if ( clsDestination == String. class )
			return strSource;
		else if ( clsDestination == java.sql.Time. class )
		{
			// TODO: Time. valueOf ( "hh:mm:ss" ) instead
			SimpleDateFormat			sdFormat = new SimpleDateFormat ( "HH:mm:ss" );
			try
			{
/*System. out. println ( "Stirng -> Time : " + 
	strSource + " -> " + sdFormat. parse ( strSource ). getTime ( ) );*/
				return new Time ( sdFormat. parse ( strSource ). getTime ( ) );
			}
			catch ( ParseException pE ) { ; }
			
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"String " + strSource + " to Time" );
		}
		else if ( clsDestination == java.sql.Timestamp. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"String to Timestamp" );

		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convert: Do not know how to convert String to " +
							clsDestination. getName ( ) );
	} // end of static Object convert ( String strSource, Class clsDestination ) throws SQLException

	static Object convert ( Boolean bSource, Class clsDestination ) throws SQLException
	{
		if ( bSource == null )
			return null;
			
		if ( clsDestination == InputStream. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Boolean to InputStream" );
		else if ( clsDestination == BigDecimal. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Boolean to BigDecimal" );
		else if ( clsDestination == Boolean. class )
			return bSource;
		else if ( clsDestination == Byte. class )
		{
			if ( bSource. booleanValue ( ) )
				return new Byte ( ( byte ) 1 );
			else
				return new Byte ( ( byte ) 0 );
		}
		else if ( clsDestination == byte [ ]. class )
		{
			byte [ ]		szbtResult = new byte [ 1 ];
			if ( bSource. booleanValue ( ) )
				szbtResult [ 0 ] = 1;
			else
				szbtResult [ 0 ] = 0;
				
			return szbtResult;
		}
		else if ( clsDestination == Reader. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Boolean to Reader" );
		else if ( clsDestination == java.sql.Date. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Boolean to Date" );
		
		int		nValue = 0;
		if ( bSource. booleanValue ( ) )
			nValue = 1;

		if ( clsDestination == Double. class )
			return new Double ( nValue );
		else if ( clsDestination == Float. class )
			return new Float ( nValue );
		else if ( clsDestination == Integer. class )
			return new Integer ( nValue );
		else if ( clsDestination == Long. class )
			return new Long ( nValue );
		else if ( clsDestination == Short. class )
			return new Short ( ( short ) nValue );
		
		if ( clsDestination == Object. class )
			return bSource;
		else if ( clsDestination == String. class )
		{
			if ( bSource. booleanValue ( ) )
				return "true";
			else
				return "false";
		}
		else if ( clsDestination == java.sql.Time. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Boolean to Time" );
		else if ( clsDestination == java.sql.Timestamp. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Boolean to Timestamp" );

		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convert: Do not know how to convert Boolean to " +
							clsDestination. getName ( ) );
	} // end of static Object convert ( Boolean bSource, Class clsDestination ) throws SQLException
	
	static Object convert ( Byte btSource, Class clsDestination ) throws SQLException
	{
		if ( btSource == null )
			return null;
			
		if ( clsDestination == InputStream. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Byte to InputStream" );
		else if ( clsDestination == BigDecimal. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Byte to BigDecimal" );
		else if ( clsDestination == Boolean. class )
		{
			if ( btSource. byteValue ( ) == 0 )
				return new Boolean ( false );
			else
				return new Boolean ( true );
		}
		else if ( clsDestination == Byte. class )
			return btSource;
		else if ( clsDestination == byte [ ]. class )
			return new byte [ ] { btSource. byteValue ( ) };
		else if ( clsDestination == Reader. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Byte to Reader" );
		else if ( clsDestination == java.sql.Date. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Byte to Date" );
		else if ( clsDestination == Double. class )
			return new Double ( btSource. byteValue ( ) );
		else if ( clsDestination == Float. class )
			return new Float ( btSource. byteValue ( ) );
		else if ( clsDestination == Integer. class )
			return new Integer ( btSource. byteValue ( ) );
		else if ( clsDestination == Long. class )
			return new Long ( btSource. byteValue ( ) );
		else if ( clsDestination == Short. class )
			return new Short ( btSource. byteValue ( ) );
		else if ( clsDestination == Object. class )
			return btSource;
		else if ( clsDestination == String. class )
			return btSource. toString ( );
		else if ( clsDestination == java.sql.Time. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Byte to Time" );
		else if ( clsDestination == java.sql.Timestamp. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Byte to Timestamp" );

		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convert: Do not know how to convert Byte to " +
							clsDestination. getName ( ) );
	} // end of static Object convert ( Byte btSource, Class clsDestination ) throws SQLException

	static Object convert ( byte [ ] szbtSource, Class clsDestination ) throws SQLException
	{
		if ( szbtSource == null )
			return null;
			
		if ( clsDestination == byte [ ]. class )
			return szbtSource;
		else if ( clsDestination == Object. class )
			return szbtSource;
		else if ( clsDestination == String. class )
			return new String ( szbtSource );

		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convert: Do not know how to convert byte [ ] to " +
							clsDestination. getName ( ) );
	} // end of static Object convert ( byte [ ] szbtSource, Class clsDestination ) throws SQLException
	
	static Object convert ( java.sql.Date dtSource, Class clsDestination ) throws SQLException
	{
		if ( dtSource == null )
			return null;
			
		if ( clsDestination == java.sql.Date. class )
			return dtSource;
		else if ( clsDestination == Double. class )
			return new Double ( dtSource. getTime ( ) );
		else if ( clsDestination == Float. class )
			return new Float ( dtSource. getTime ( ) );
		else if ( clsDestination == Integer. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Date to Integer" );
		else if ( clsDestination == Long. class )
			return new Long ( dtSource. getTime ( ) );
		else if ( clsDestination == Short. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Date to Short" );
		else if ( clsDestination == Object. class )
			return dtSource;
		else if ( clsDestination == String. class )
			return dtSource. toString ( );
		else if ( clsDestination == java.sql.Time. class )
			return new java.sql.Time ( dtSource. getTime ( ) );
		else if ( clsDestination == java.sql.Timestamp. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Date to Timestamp" );

		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convert: Do not know how to convert Date to " +
							clsDestination. getName ( ) );
	} // end of static Object convert ( java.sql.Date dtSource, Class clsDestination ) throws SQLException

	static Object convert ( Double nSource, Class clsDestination ) throws SQLException
	{
		if ( nSource == null )
			return null;
			
		if ( clsDestination == java.sql.Date. class )
			return new java.sql.Date ( nSource. longValue ( ) );
		else if ( clsDestination == Double. class )
			return nSource;
		else if ( clsDestination == Float. class )
			return new Float ( nSource. floatValue ( ) );
		else if ( clsDestination == Integer. class )
			return new Integer ( nSource. intValue ( ) );
		else if ( clsDestination == Long. class )
			return new Long ( nSource. longValue ( ) );
		else if ( clsDestination == Short. class )
			return new Short ( nSource. shortValue ( ) );
		else if ( clsDestination == Object. class )
			return nSource;
		else if ( clsDestination == String. class )
			return nSource. toString ( );
		else if ( clsDestination == java.sql.Time. class )
			return new java.sql.Time ( nSource. longValue ( ) );

		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convert: Do not know how to convert Double to " +
							clsDestination. getName ( ) );
	} // end of static Object convert ( Double nSource, Class clsDestination ) throws SQLException

	static Object convert ( Float nSource, Class clsDestination ) throws SQLException
	{
		if ( nSource == null )
			return null;
			
		if ( clsDestination == java.sql.Date. class )
			return new java.sql.Date ( nSource. longValue ( ) );
		else if ( clsDestination == Double. class )
			return new Double ( nSource. doubleValue ( ) );
		else if ( clsDestination == Float. class )
			return nSource;
		else if ( clsDestination == Integer. class )
			return new Integer ( nSource. intValue ( ) );
		else if ( clsDestination == Long. class )
			return new Long ( nSource. longValue ( ) );
		else if ( clsDestination == Short. class )
			return new Short ( nSource. shortValue ( ) );
		else if ( clsDestination == Object. class )
			return nSource;
		else if ( clsDestination == String. class )
			return nSource. toString ( );
		else if ( clsDestination == java.sql.Time. class )
			return new java.sql.Time ( nSource. longValue ( ) );

		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convert: Do not know how to convert Float to " +
							clsDestination. getName ( ) );
	} // end of static Object convert ( Float nSource, Class clsDestination ) throws SQLException

	static Object convert ( Integer nSource, Class clsDestination ) throws SQLException
	{
		if ( nSource == null )
			return null;
			
		if ( clsDestination == java.sql.Date. class )
			return new java.sql.Date ( nSource. longValue ( ) );
		else if ( clsDestination == Double. class )
			return new Double ( nSource. doubleValue ( ) );
		else if ( clsDestination == Float. class )
			return new Float ( nSource. floatValue ( ) );
		else if ( clsDestination == Integer. class )
			return nSource;
		else if ( clsDestination == Long. class )
			return new Long ( nSource. longValue ( ) );
		else if ( clsDestination == Short. class )
			return new Short ( nSource. shortValue ( ) );
		else if ( clsDestination == Object. class )
			return nSource;
		else if ( clsDestination == String. class )
			return nSource. toString ( );
		else if ( clsDestination == java.sql.Time. class )
			return new java.sql.Time ( nSource. longValue ( ) );

		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convert: Do not know how to convert Integer to " +
							clsDestination. getName ( ) );
	} // end of static Object convert ( Integer nSource, Class clsDestination ) throws SQLException

	static Object convert ( Long nSource, Class clsDestination ) throws SQLException
	{
		if ( nSource == null )
			return null;
			
		if ( clsDestination == java.sql.Date. class )
			return new java.sql.Date ( nSource. longValue ( ) );
		else if ( clsDestination == Double. class )
			return new Double ( nSource. doubleValue ( ) );
		else if ( clsDestination == Float. class )
			return new Float ( nSource. floatValue ( ) );
		else if ( clsDestination == Integer. class )
			return new Integer ( nSource. intValue ( ) );
		else if ( clsDestination == Long. class )
			return nSource;
		else if ( clsDestination == Short. class )
			return new Short ( nSource. shortValue ( ) );
		else if ( clsDestination == Object. class )
			return nSource;
		else if ( clsDestination == String. class )
			return nSource. toString ( );
		else if ( clsDestination == java.sql.Time. class )
			return new java.sql.Time ( nSource. longValue ( ) );

		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convert: Do not know how to convert Long to " +
							clsDestination. getName ( ) );
	} // end of static Object convert ( Long nSource, Class clsDestination ) throws SQLException

	static Object convert ( Short nSource, Class clsDestination ) throws SQLException
	{
		if ( nSource == null )
			return null;
			
		if ( clsDestination == java.sql.Date. class )
			return new java.sql.Date ( nSource. longValue ( ) );
		else if ( clsDestination == Double. class )
			return new Double ( nSource. doubleValue ( ) );
		else if ( clsDestination == Float. class )
			return new Float ( nSource. floatValue ( ) );
		else if ( clsDestination == Integer. class )
			return new Integer ( nSource. intValue ( ) );
		else if ( clsDestination == Long. class )
			return new Long ( nSource. longValue ( ) );
		else if ( clsDestination == Short. class )
			return nSource;
		else if ( clsDestination == Object. class )
			return nSource;
		else if ( clsDestination == String. class )
			return nSource. toString ( );
		else if ( clsDestination == java.sql.Time. class )
			return new java.sql.Time ( nSource. longValue ( ) );

		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convert: Do not know how to convert Short to " +
							clsDestination. getName ( ) );
	} // end of static Object convert ( Short nSource, Class clsDestination ) throws SQLException

	static Object convert ( Time tSource, Class clsDestination ) throws SQLException
	{
		if ( tSource == null )
			return null;
			
		if ( clsDestination == java.sql.Date. class )
			return new java.sql.Date ( tSource. getTime ( ) );
		else if ( clsDestination == Double. class )
			return new Double ( tSource. getTime ( ) );
		else if ( clsDestination == Float. class )
			return new Float ( tSource. getTime ( ) );
		else if ( clsDestination == Integer. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Time to Integer" );
		else if ( clsDestination == Long. class )
			return new Long ( tSource. getTime ( ) );
		else if ( clsDestination == Short. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Time to Short" );
		else if ( clsDestination == Object. class )
			return tSource;
		else if ( clsDestination == String. class )
			return tSource. toString ( );
		else if ( clsDestination == java.sql.Time. class )
			return tSource;
		else if ( clsDestination == java.sql.Timestamp. class )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::convert: Do not know how to convert " +
								"Time to Timestamp" );

		throw new SQLException (
							"com.fourd.jdbc.SQLTools::convert: Do not know how to convert Time to " +
							clsDestination. getName ( ) );
	} // end of static Object convert ( Time dtSource, Class clsDestination ) throws SQLException
	
	/*
		Returns java.sql.Types constant that corresponds to 4D type in n4DType
	*/
	static int map4DTypeToSQLType ( int n4DType ) throws SQLException
	{
		if ( n4DType == opConstants. ALPHANUMERIC )
			return Types. VARCHAR;
		else if ( n4DType == opConstants. BOOLEAN )
			return Types. BIT;
		else if ( n4DType == opConstants. INTEGER )
			return Types. SMALLINT;
		else if ( n4DType == opConstants. LONGINTEGER )
			return Types. INTEGER;
		else if ( n4DType == opConstants. REAL )
			return Types. DOUBLE;
		else if ( n4DType == opConstants. TEXT )
			return Types. LONGVARCHAR;
		else if ( n4DType == opConstants. TIME )
			return Types. TIME;
		else if ( n4DType == opConstants. DATE )
			return Types. DATE;
		else if ( n4DType == opConstants. BLOB )
			return Types. BLOB;
		else if ( n4DType == opConstants. PICTURE )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::map4DTypeToSQLType: " +
								"Unimplemented data type - PICTURE" );
		else if ( n4DType == opConstants. SUBTABLE )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::map4DTypeToSQLType: " +
								"Unimplemented data type - SUBTABLE" );
		else if ( n4DType == opConstants. ALPHANUMERIC2 )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::map4DTypeToSQLType: " +
								"Unimplemented data type - ALPHANUMERIC2" );
			
		throw new SQLException (
							"com.fourd.jdbc.SQLTools::map4DTypeToSQLType: " +
							"Unknown data type - " + n4DType );
	}

	/*
		Returns a 4D name of a 4D type
	*/
	static String get4DTypeName ( int n4DType ) throws SQLException
	{
		if ( n4DType == opConstants. ALPHANUMERIC )
			return "ALPHANUMERIC";
		else if ( n4DType == opConstants. BOOLEAN )
			return "BOOLEAN";
		else if ( n4DType == opConstants. INTEGER )
			return "INTEGER";
		else if ( n4DType == opConstants. LONGINTEGER )
			return "LONGINTEGER";
		else if ( n4DType == opConstants. REAL )
			return "REAL";
		else if ( n4DType == opConstants. TEXT )
			return "TEXT";
		else if ( n4DType == opConstants. TIME )
			return "TIME";
		else if ( n4DType == opConstants. DATE )
			return "DATE";
		else if ( n4DType == opConstants. BLOB )
			return "BLOB";
		else if ( n4DType == opConstants. PICTURE )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::get4DTypeName: " +
								"Unimplemented data type - PICTURE" );
		else if ( n4DType == opConstants. SUBTABLE )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::get4DTypeName: " +
								"Unimplemented data type - SUBTABLE" );
		else if ( n4DType == opConstants. ALPHANUMERIC2 )
			throw new SQLException (
								"com.fourd.jdbc.SQLTools::get4DTypeName: " +
								"Unimplemented data type - ALPHANUMERIC2" );
			
		throw new SQLException (
							"com.fourd.jdbc.SQLTools::get4DTypeName: " +
							"Unknown data type - " + n4DType );
	}

	/*
		Returns 4D type constant that corresponds to java.sql.Types in nSQLType
	*/
	static int mapSQLTypeTo4DType ( int nSQLType ) throws SQLException
	{
		if ( nSQLType == Types. VARCHAR )
			return opConstants. ALPHANUMERIC;
		else if ( nSQLType == Types. BIT )
			return opConstants. BOOLEAN;
		else if ( nSQLType == Types. SMALLINT )
			return opConstants. INTEGER;
		else if ( nSQLType == Types. INTEGER )
			return opConstants. LONGINTEGER;
		else if ( nSQLType == Types. DOUBLE )
			return opConstants. REAL;
		else if ( nSQLType == Types. LONGVARCHAR )
			return opConstants. TEXT;
		else if ( nSQLType == Types. TIME )
			return opConstants. TIME;
		else if ( nSQLType == Types. DATE )
			return opConstants. DATE;
		else if ( nSQLType == Types. BLOB )
			return opConstants. BLOB;
			
		throw new SQLException (
							"com.fourd.jdbc.SQLTools::mapSQLTypeTo4DType: " +
							"Unknown/unsupported SQL data type - " + nSQLType );
	}

	/*
		Returns 4D type constant that corresponds to java class
	*/
	static int mapJavaClassTo4DType ( Object obj ) throws SQLException
	{
		if ( obj. getClass ( ) == String. class )
			return opConstants. ALPHANUMERIC;
		else if ( obj. getClass ( ) == Double. class )
			return opConstants. REAL;
		else if ( obj. getClass ( ) == Short. class )
			return opConstants. INTEGER;
		else if ( obj. getClass ( ) == Integer. class )
			return opConstants. LONGINTEGER;
		else if ( obj. getClass ( ) == java.sql.Date. class )
			return opConstants. DATE;
		else if ( obj. getClass ( ) == java.sql.Time. class )
			return opConstants. TIME;
		else if ( obj. getClass ( ) == Boolean. class )
			return opConstants. BOOLEAN;
		else if ( obj. getClass ( ) == byte [ ]. class )
			return opConstants. BLOB;
		/*else if ( obj. getClass ( ) == String )
			return opConstants. TEXT;*/
			
		throw new SQLException (
							"com.fourd.jdbc.SQLTools::mapJavaClassTo4DType: " +
							"Unknown/unsupported java class - " + obj. getClass ( ). getName ( ) );
	}
}